a {
  box-shadow: 0 0 currentColor;
}

b {
  box-shadow: 0 0 currentColor inset;
}

c {
  box-shadow: 0 0 rgb(255,0,0) inset;
}

d {
  box-shadow: 0 0 rgb(255,0,0);
}

e {
  box-shadow: 0 0 rgb(255,0,0) inset;
}

f {
  box-shadow: 0 0 currentColor inset;
}

g {
  box-shadow: 0 0 rgb(255,0,0) inset;
}

h {
  box-shadow: 0 0 rgb(255,0,0) inset;
}

i {
  box-shadow: 0 0 rgb(255,0,0);
}

j {
  box-shadow: 0 0 rgb(255,0,0) inset;
}

k {
  box-shadow: 0 0 rgb(255,0,0) inset;
}
