a {
  border-image-width: 1 2 3 4;
}

b {
  border-image-width: 1 2 3;
}

c {
  border-image-width: 1 2;
}

d {
  border-image-width: 1;
}

e {
  border-image-width: 0;
}

f {
  border-image-width: 1px 2px 3 4px;
}

n {
  border-image-width: 1 2;
}
