* {
  all: unset;
}

.background {
  background-color: red;
}

button {
  padding: 10px;
}

#topleft {
  border-top-left-radius: 20px;
  border-left: 5px solid blue;
  border-top: 5px solid blue;
}

#topright {
  border-top-right-radius: 20px;
  border-right: 5px solid blue;
  border-top: 5px solid blue;
}

#bottomleft {
  border-bottom-left-radius: 20px;
  border-left: 5px solid blue;
  border-bottom: 5px solid blue;
}

#bottomright {
  border-bottom-right-radius: 20px;
  border-right: 5px solid blue;
  border-bottom: 5px solid blue;
}

#reference {
  border: 5px solid blue;
  border-radius: 20px;
}
