frame.front {
  border: 2px solid white;
  background-color: rgba(228, 0, 0, 0.8);
}

frame.back {
  border: 2px solid white;
  background-color: rgba(228, 0, 0, 0.8);
}

frame.right {
  border: 2px solid white;
  background-color: rgba(127, 231, 25, 0.8);
}

frame.left {
  border: 2px solid white;
  background-color: rgba(127, 231, 25, 0.8);
}

frame.top {
  border: 2px solid white;
  background-color: rgba(114, 159, 207, 0.8);
}

frame.bottom {
  border: 2px solid white;
  background-color: rgba(114, 159, 207, 0.8);
}
