window {
  background: repeating-radial-gradient(red 50%, blue, red);
}

#reference {
  background-image: radial-gradient(red, blue, red), repeating-radial-gradient(red 50%, blue, red);
  background-size: 50% 50%, 100% 100%;
  background-repeat: no-repeat;
  background-position: center;
}

