N_("Select an Object");
N_("Show Details");
N_("Show all Objects");
N_("Collect Statistics");
N_("Show Details");
N_("Show all Resources");
N_("Miscellaneous");
N_("Properties");
N_("Signals");
N_("Child Properties");
N_("Hierarchy");
N_("Selector");
N_("Style Classes");
N_("Style Properties");
N_("CSS nodes");
N_("CSS");
N_("Size Groups");
N_("Data");
N_("Actions");
N_("Menu");
N_("Gestures");
N_("Magnifier");
N_("Objects");
N_("Statistics");
N_("Resources");
N_("CSS");
N_("Visual");
N_("General");
