@import "reset-to-defaults.css";

window {
  background-color: black;
}

#reference {
  background: linear-gradient(red 50%, yellow 50%);
}

#red {
  background-color: red;
}

#yellow {
  background-color: yellow;
}
