/* You can edit the text in this window to change the
 * appearance of this Window.
 * Be careful, if you screw it up, nothing might be visible
 * anymore. :)
 */

/* This CSS resets all properties to their defaults values
 *    and overrides all user settings and the theme in use */
@import url("resource://css_basics/reset.css");

/* Set a very futuristic style by default */
* {
  color: green;
  font-family: Monospace;
  border: 1px solid;
}

window {
  background-color: white;
}

/* Make sure selections are visible */
selection {
  background-color: darkGreen;
  color: black;
}
