/* Dummy header for the Quartz backend. */

typedef int NSTrackingRectTag;

@interface NSAutoreleasePool {}
@end

@interface NSCursor {}
@end

@interface NSEvent {}
@end

@interface NSImage {}
@end

@interface NSView {}
@end

@interface NSWindow {}
@end

@interface NSString {}
@end

@protocol NSDraggingInfo
@end
