* {}
E {}
E F {}
E > F {}
E + F {}
E#id {}
#id {}
tab:first-child {}
tab:last-child {}
tab:first-child {}
tab:last-child {}
tab:nth-child(even) {}
tab:nth-child(odd) {}
.some-class {}
.some-class.another-class {}
.some-class .another-class {}
E * {}
E .class {}
E > .foo {}
E > #id {}
E:active {}
E:hover {}
E:selected {}
E:disabled {}
E:indeterminate {}
E:focus {}
E:active:hover {}
* > .notebook tab:first-child .label:focus {}
E, F {}
E, F /* comment here */ {}
E,/* comment here */ F {}
E1.e1_2 #T3_4 {}
E:first-child {}
E:last-child {}
E:first-child {}
E:last-child {}
E:nth-child(even) {}
E:nth-child(odd) {}
E:focus tab {}
