#ifndef GTKDOC_TESTER_H
#define GTKDOC_TESTER_H

#include <glib.h>

// FIXME: this should not be needed, if we have a title + long-desc
/**
 * FOO:
 *
 * Placeholder to have non empty docs.
 */
#define FOO 1

#endif // GTKDOC_TESTER_H
