N_("Address");
N_("Reference count");
N_("State");
N_("Buildable ID");
N_("Default Widget");
N_("Properties");
N_("Focus Widget");
N_("Properties");
N_("Mnemonic Label");
N_("Allocated size");
N_("Clip area");
N_("Tick callback");
N_("Frame count");
N_("Frame rate");
N_("Accessible role");
N_("Accessible name");
N_("Accessible description");
N_("Mapped");
N_("Realized");
N_("Is Toplevel");
N_("Child Visible");
