/* Dummy header for the DirectFB backend. */

typedef int DFBColor;
typedef int DFBSurfaceCapabilities;
typedef int DFBSurfacePixelFormat;
typedef int DFBWindowCapabilities;
typedef int DFBWindowOptions;

typedef struct IDirectFBWindow IDirectFBWindow;
typedef struct IDirectFBSurface IDirectFBSurface;
