@import "reset-to-defaults.css";

window {
  background-color: black;
}

box {
  background-color: @doesntexist;
  box-shadow: @nonexistent 5px 5px;
  border-image: linear-gradient(to top, @notaavailable, @whatever);
}

#reference {
  background-color: transparent;
  box-shadow: none;
  border-image: none;
}
