a {
  background-blend-mode: initial;
}

b {
  background-blend-mode: inherit;
}

c {
  background-blend-mode: unset;
}

d {
  background-blend-mode: normal;
}

e {
  background-blend-mode: color-burn, color-dodge, color, darken, difference, exclusion, hard-light, hue, lighten, luminosity, multiply, normal, overlay, saturation, screen, soft-light;
}
