a {
  margin-left: calc(3px + 1em);
}

b {
  transition-duration: calc(1s - 100ms + -100ms);
}

c {
  margin-left: calc(-2px + 1em +    4px);
}

d {
  background-size: calc( 100% - 10px );
}

e {
  border-left-width: calc(1px + 1px);
}

f {
  background-size: calc(2 * 3px + 4px * 5 - 6px / 3);
}

g {
  background-size: calc((((((1px))))));
}

h {
  background-size: calc(3 * ((4px + ((4 * (6px + 3em)) / 3)) + 5em));
}

i {
  border-top-left-radius: calc(1px * (3 + 2)) calc(1px / ( 3 / (2 * 1.5)));
}

j {
  border-bottom-right-radius: calc((1 + 1 + (1 + 1) + 1) * 1px) calc((1 / 1) * 1px);
}

k {
  margin-left: calc(calc(1px));
}
