/* Dummy header for the Xlib backend. */

typedef unsigned long CARD32;

typedef CARD32 XID;

typedef XID Atom;
typedef XID Colormap;
typedef XID Cursor;
typedef XID Window;
typedef XID VisualID;

typedef struct _Display Display;
typedef struct _GC GC;
typedef struct _Screen Screen;
typedef struct _Visual Visual;
typedef struct _XImage XImage;
