a {
  background-clip: border-box, content-box, padding-box, border-box, border-box, padding-box, padding-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none;
  background-origin: border-box, content-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: left top, left top, 10%, left top, center, 5px bottom, left top, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, 10%;
}

b {
  background-clip: border-box, content-box, content-box, content-box, border-box, padding-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */, none;
  background-origin: border-box, content-box, padding-box, content-box, border-box, content-box, border-box, border-box, border-box;
  background-position: 10%, 10%, 5px bottom, 10%, left top, left top, left top, center, center;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, 10%, cover, 10%, auto, auto, auto;
}

c {
  background-clip: padding-box, border-box, padding-box, padding-box, padding-box, content-box, border-box, border-box;
  background-color: @home;
  background-image: none, none, none, none, none /* FIXME */, none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, padding-box, content-box, content-box, padding-box, border-box;
  background-position: left top, center, left top, left top, left top, left top, 10%, center;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, cover, auto, auto, auto;
}

d {
  background-clip: border-box, content-box, content-box, padding-box, padding-box, padding-box, content-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none, none;
  background-origin: padding-box, content-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box;
  background-position: left top, 10%, left top, 10%, 10%, left top, 10%, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, cover, auto;
}

e {
  background-clip: border-box, content-box, border-box, padding-box, padding-box, border-box, padding-box, border-box, content-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none, none, none;
  background-origin: padding-box, content-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, border-box, padding-box;
  background-position: 10%, 10%, 5px bottom, 5px bottom, left top, center, left top, 10%, left top, center;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, cover, auto, auto, auto, 10%;
}

f {
  background-clip: border-box, border-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none;
  background-origin: content-box, border-box, content-box;
  background-position: left top, left top, left top;
  background-repeat: round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto;
}

g {
  background-clip: border-box, content-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, content-box, content-box;
  background-position: 5px bottom, left top, center;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: 10%, auto, cover;
}

h {
  background-clip: content-box, padding-box, padding-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */;
  background-origin: content-box, content-box, padding-box, padding-box, padding-box, border-box;
  background-position: center, 10%, center, left top, 5px bottom, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: 10%, auto, cover, auto, auto, auto;
}

i {
  background-clip: padding-box, content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none;
  background-origin: padding-box, content-box;
  background-position: left top, left top;
  background-repeat: repeat, round no-repeat;
  background-size: auto, auto;
}

j {
  background-clip: padding-box, border-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none;
  background-origin: padding-box, border-box, padding-box, padding-box;
  background-position: left top, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto;
}

k {
  background-clip: content-box, padding-box, content-box, border-box, padding-box, border-box, padding-box, content-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none;
  background-origin: content-box, border-box, content-box, border-box, content-box, padding-box, padding-box, content-box;
  background-position: left top, left top, center, left top, left top, left top, left top, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: 10%, auto, 10%, cover, auto, auto, auto, auto;
}

l {
  background-clip: border-box, content-box, content-box, padding-box;
  background-color: @home;
  background-image: none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, content-box, padding-box;
  background-position: left top, left top, 10%, 10%;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

m {
  background-clip: padding-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box;
  background-position: center, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

n {
  background-clip: border-box, content-box, content-box, border-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, content-box, padding-box, content-box, padding-box, padding-box;
  background-position: 10%, left top, center, left top, 10%, left top, left top;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, auto, 10%, auto;
}

o {
  background-clip: content-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box;
  background-position: 5px bottom, left top;
  background-repeat: round no-repeat, repeat;
  background-size: auto, 10%;
}

p {
  background-clip: content-box;
  background-color: currentColor;
  background-image: none;
  background-origin: content-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

q {
  background-clip: border-box, padding-box, content-box, border-box, padding-box, border-box, border-box, border-box, content-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none /* FIXME */;
  background-origin: border-box, padding-box, content-box, padding-box, padding-box, content-box, padding-box, border-box, content-box, content-box;
  background-position: 10%, 10%, left top, left top, 10%, left top, 5px bottom, left top, center, center;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, cover, auto, auto, cover, auto, auto, auto, auto, auto;
}

r {
  background-clip: content-box, border-box, content-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, border-box, content-box, padding-box;
  background-position: center, 10%, left top, 5px bottom;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

s {
  background-clip: padding-box, content-box, content-box, border-box, border-box, border-box, padding-box, content-box, padding-box, content-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: none, none, none, none, none /* FIXME */, none /* FIXME */, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box, padding-box, padding-box, border-box;
  background-position: 10%, left top, left top, left top, left top, center, 5px bottom, center, center, 5px bottom, 10%, 5px bottom, center;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

t {
  background-clip: padding-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none;
  background-origin: padding-box, padding-box;
  background-position: 5px bottom, left top;
  background-repeat: round no-repeat, round no-repeat;
  background-size: 10%, auto;
}

u {
  background-clip: border-box, border-box;
  background-color: currentColor;
  background-image: none, none;
  background-origin: padding-box, border-box;
  background-position: left top, left top;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

v {
  background-clip: content-box, content-box, border-box, padding-box, content-box, padding-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none /* FIXME */, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, padding-box, padding-box, content-box, padding-box, content-box;
  background-position: left top, 5px bottom, center, center, 5px bottom, 10%, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, 10%, auto, auto, auto, auto;
}

w {
  background-clip: border-box, content-box, border-box, content-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box, border-box, border-box, border-box, border-box, content-box, border-box, border-box, border-box, padding-box, border-box, padding-box, border-box, border-box, content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none /* FIXME */, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none;
  background-origin: padding-box, content-box, border-box, content-box, border-box, border-box, padding-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box, border-box, padding-box, content-box;
  background-position: left top, left top, 10%, center, 10%, 10%, left top, 5px bottom, left top, left top, left top, left top, left top, left top, left top, left top, 10%, 10%, center, center, left top, left top, left top, left top, center;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, 10%;
}

x {
  background-clip: content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none;
  background-origin: content-box, padding-box;
  background-position: 10%, left top;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

y {
  background-clip: border-box, border-box, content-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, content-box, padding-box;
  background-position: 5px bottom, left top, 10%, 5px bottom;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat;
  background-size: cover, auto, cover, auto;
}

z {
  background-clip: content-box, content-box, content-box, border-box, padding-box, padding-box, border-box, padding-box, padding-box, border-box, border-box, padding-box, padding-box, content-box, padding-box, content-box, content-box, border-box, padding-box, padding-box, padding-box, border-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none;
  background-origin: content-box, content-box, content-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box, border-box, border-box, padding-box, padding-box, border-box, padding-box, content-box, content-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, left top, left top, center, left top, 5px bottom, left top, 10%, 10%, center, 10%, left top, left top, 5px bottom, left top, 5px bottom, 10%, 5px bottom, 10%, left top, 10%, left top, center;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

ab {
  background-clip: padding-box, padding-box, border-box, border-box, content-box, content-box, border-box, border-box, border-box, content-box, border-box, border-box, border-box, border-box, padding-box, content-box, content-box, content-box, border-box, content-box, border-box, padding-box, border-box, border-box, content-box, border-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, border-box, padding-box, padding-box, content-box, border-box, padding-box, padding-box, content-box, padding-box, border-box, border-box, border-box, content-box, padding-box, content-box, content-box, padding-box, content-box, border-box, padding-box, padding-box, border-box, content-box, border-box, content-box;
  background-position: 5px bottom, left top, center, 5px bottom, left top, 5px bottom, 5px bottom, left top, left top, left top, center, left top, 10%, center, center, 10%, center, left top, left top, 10%, center, 5px bottom, 5px bottom, left top, left top, center, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto;
}

bb {
  background-clip: content-box, padding-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box, padding-box;
  background-position: 5px bottom, center, 5px bottom;
  background-repeat: repeat, repeat, round no-repeat;
  background-size: 10%, auto, auto;
}

cb {
  background-clip: border-box, padding-box, border-box, border-box, border-box, border-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none;
  background-origin: padding-box, padding-box, border-box, padding-box, border-box, border-box, padding-box;
  background-position: left top, center, center, center, left top, 10%, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, cover, auto, cover, auto;
}

db {
  background-clip: padding-box;
  background-color: @home;
  background-image: none;
  background-origin: border-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

eb {
  background-clip: border-box, border-box, padding-box, content-box, border-box, padding-box, padding-box, border-box, padding-box, padding-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none, none /* FIXME */, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, border-box, border-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box;
  background-position: left top, left top, center, 5px bottom, left top, 10%, 10%, left top, left top, center;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, 10%, auto, auto, auto;
}

fb {
  background-clip: border-box, padding-box, content-box, border-box, padding-box, content-box, border-box, border-box, border-box, padding-box, border-box, border-box, border-box, content-box, border-box, border-box, border-box, border-box, content-box, border-box, border-box, padding-box, content-box, border-box, border-box, padding-box, padding-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none /* FIXME */, none /* FIXME */, none, none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, content-box, border-box, padding-box, content-box, border-box, padding-box, padding-box, border-box, padding-box, padding-box, content-box, content-box, border-box, border-box, padding-box, padding-box, content-box, border-box, padding-box, padding-box, content-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 10%, center, 5px bottom, 5px bottom, left top, center, left top, 5px bottom, left top, left top, 5px bottom, 5px bottom, left top, left top, 5px bottom, 10%, 10%, 5px bottom, 5px bottom, 5px bottom, left top, 10%, left top, 10%, center, 5px bottom, left top, 10%, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, 10%, auto, cover, auto, auto, auto;
}

gb {
  background-clip: padding-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

hb {
  background-clip: padding-box, border-box, content-box, content-box, padding-box, padding-box, content-box, content-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none;
  background-origin: border-box, padding-box, content-box, content-box, padding-box, padding-box, content-box, content-box, padding-box;
  background-position: center, 10%, 5px bottom, 5px bottom, 5px bottom, center, left top, 5px bottom, left top;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

ib {
  background-clip: border-box, padding-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box;
  background-position: 5px bottom, 5px bottom;
  background-repeat: repeat, round no-repeat;
  background-size: 10%, auto;
}

jb {
  background-clip: border-box, border-box, border-box, padding-box, border-box, border-box, border-box, border-box, padding-box, content-box, border-box, border-box, padding-box, padding-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, padding-box, border-box, padding-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box, padding-box, padding-box, content-box;
  background-position: 10%, 10%, center, left top, left top, 5px bottom, 10%, 10%, left top, 5px bottom, left top, 5px bottom, 10%, left top, 5px bottom;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto, cover, auto, auto;
}

kb {
  background-clip: border-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */;
  background-origin: padding-box, padding-box;
  background-position: 5px bottom, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

lb {
  background-clip: content-box, border-box, padding-box, border-box, border-box, content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: content-box, border-box, content-box, padding-box, border-box, content-box, padding-box;
  background-position: 5px bottom, left top, center, left top, 10%, center, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat;
  background-size: 10%, auto, auto, auto, auto, auto, auto;
}

mb {
  background-clip: content-box, border-box, border-box, border-box, padding-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none;
  background-origin: content-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: left top, left top, 5px bottom, center, center, left top, 10%;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

nb {
  background-clip: border-box, content-box, padding-box;
  background-color: @home;
  background-image: none /* FIXME */, none, none;
  background-origin: padding-box, content-box, padding-box;
  background-position: left top, left top, left top;
  background-repeat: repeat, repeat, round no-repeat;
  background-size: auto, auto, auto;
}

ob {
  background-clip: padding-box, border-box, border-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, border-box, content-box, padding-box, border-box, padding-box, content-box, border-box;
  background-position: center, left top, left top, left top, left top, left top, center, 5px bottom, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, cover, auto, 10%;
}

pb {
  background-clip: border-box, border-box, border-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box;
  background-position: left top, 10%, center;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: auto, auto, auto;
}

qb {
  background-clip: border-box, content-box, content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, border-box, border-box, border-box;
  background-position: left top, center, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto;
}

rb {
  background-clip: padding-box, content-box, border-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: border-box, content-box, padding-box, content-box;
  background-position: center, 10%, 5px bottom, center;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto;
}

sb {
  background-clip: border-box, padding-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none;
  background-origin: padding-box, content-box, padding-box;
  background-position: left top, 5px bottom, 5px bottom;
  background-repeat: round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto;
}

tb {
  background-clip: border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: padding-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: auto;
}

ub {
  background-clip: padding-box, border-box, border-box, border-box, content-box, content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, border-box, padding-box, content-box, content-box;
  background-position: left top, center, 10%, left top, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, 10%;
}

vb {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: round no-repeat;
  background-size: auto;
}

wb {
  background-clip: padding-box, border-box, content-box, padding-box, padding-box, padding-box, padding-box;
  background-color: @home;
  background-image: none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box, border-box, padding-box, content-box, padding-box, padding-box;
  background-position: left top, left top, center, left top, 10%, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

xb {
  background-clip: border-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */;
  background-origin: border-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: cover;
}

yb {
  background-clip: padding-box, padding-box, content-box, content-box, border-box, border-box, border-box, content-box, padding-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none /* FIXME */, none;
  background-origin: padding-box, padding-box, content-box, content-box, border-box, padding-box, padding-box, content-box, padding-box, border-box, content-box, border-box;
  background-position: left top, 10%, left top, center, left top, 10%, left top, left top, 10%, left top, center, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, cover, 10%, auto, auto;
}

zb {
  background-clip: content-box, content-box, content-box, border-box, border-box, border-box, border-box, content-box, content-box, border-box, padding-box, padding-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, padding-box, border-box, border-box, border-box, border-box, content-box, content-box, border-box, padding-box, padding-box, border-box;
  background-position: left top, 10%, 10%, left top, left top, left top, left top, 5px bottom, 5px bottom, 5px bottom, 5px bottom, 10%, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, cover, 10%, auto, cover, auto, auto, auto;
}

ac {
  background-clip: padding-box, border-box, content-box, border-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, content-box, padding-box, content-box;
  background-position: left top, left top, left top, 5px bottom, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, 10%, auto, 10%, auto;
}

bc {
  background-clip: border-box, border-box, border-box;
  background-color: @home;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box;
  background-position: left top, left top, 10%;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, 10%;
}

cc {
  background-clip: border-box, content-box, border-box, content-box, content-box;
  background-color: initial;
  background-image: none, none, none, none, none;
  background-origin: border-box, border-box, padding-box, content-box, padding-box;
  background-position: left top, center, center, 5px bottom, center;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, cover, auto;
}

dc {
  background-clip: content-box, content-box, border-box, content-box, padding-box, content-box, padding-box, border-box, padding-box, padding-box;
  background-color: @home;
  background-image: none, none, none /* FIXME */, none, none, none, none /* FIXME */, none, none, none /* FIXME */;
  background-origin: content-box, content-box, border-box, border-box, padding-box, content-box, padding-box, padding-box, padding-box, border-box;
  background-position: 5px bottom, center, 10%, left top, 10%, 5px bottom, 5px bottom, left top, center, center;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto;
}

ec {
  background-clip: content-box;
  background-color: @home;
  background-image: none;
  background-origin: content-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

fc {
  background-clip: padding-box, content-box, padding-box, border-box, content-box, border-box, border-box, border-box, border-box, padding-box, border-box, content-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, none, none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, padding-box, border-box, content-box, padding-box, border-box, padding-box, content-box, padding-box, padding-box, content-box;
  background-position: center, 5px bottom, 10%, center, center, 10%, left top, left top, left top, left top, center, center;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, 10%, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto;
}

gc {
  background-clip: border-box, padding-box, content-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: padding-box, padding-box, content-box, padding-box;
  background-position: center, 5px bottom, left top, 10%;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

hc {
  background-clip: border-box, padding-box, border-box, content-box, padding-box, border-box;
  background-color: @home;
  background-image: none /* FIXME */, none, none, none /* FIXME */, none /* FIXME */, none;
  background-origin: border-box, padding-box, padding-box, content-box, border-box, padding-box;
  background-position: center, 5px bottom, left top, 10%, center, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, cover, auto, 10%;
}

ic {
  background-clip: padding-box, padding-box, content-box, border-box, border-box, content-box, border-box, content-box, content-box, padding-box, padding-box, content-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, none, none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, content-box, border-box, padding-box, content-box, border-box, content-box, content-box, padding-box, padding-box, content-box, content-box;
  background-position: left top, center, left top, 10%, 10%, 5px bottom, 5px bottom, left top, left top, 5px bottom, 10%, left top, center;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, cover, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

jc {
  background-clip: border-box, border-box;
  background-color: @home;
  background-image: none /* FIXME */, none /* FIXME */;
  background-origin: border-box, border-box;
  background-position: left top, 10%;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

kc {
  background-clip: padding-box;
  background-color: currentColor;
  background-image: none /* FIXME */;
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

lc {
  background-clip: content-box, border-box, padding-box, border-box, border-box, content-box, border-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box;
  background-position: left top, center, left top, 5px bottom, center, center, left top, center;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, cover, auto, auto, cover, auto, auto, auto;
}

mc {
  background-clip: padding-box, border-box, border-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, border-box;
  background-position: 10%, center, left top, left top, left top, 5px bottom;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, cover, auto;
}

nc {
  background-clip: content-box, border-box, content-box, padding-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, none, none /* FIXME */;
  background-origin: content-box, padding-box, content-box, padding-box;
  background-position: 5px bottom, left top, left top, 10%;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: auto, cover, auto, auto;
}

oc {
  background-clip: padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: padding-box;
  background-position: center;
  background-repeat: repeat;
  background-size: auto;
}

pc {
  background-clip: border-box, content-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, content-box, padding-box;
  background-position: left top, 5px bottom, center;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: auto, auto, auto;
}

qc {
  background-clip: content-box, content-box, border-box;
  background-color: currentColor;
  background-image: none, none, none;
  background-origin: content-box, content-box, padding-box;
  background-position: 5px bottom, center, 10%;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: auto, 10%, auto;
}

rc {
  background-clip: padding-box, padding-box, padding-box, padding-box, border-box, border-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, border-box, content-box;
  background-position: left top, 5px bottom, 10%, left top, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, cover, auto, auto, auto, auto;
}

sc {
  background-clip: content-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none, none, none;
  background-origin: content-box, border-box, border-box, padding-box, padding-box, padding-box, border-box, padding-box;
  background-position: left top, 10%, left top, 5px bottom, left top, 5px bottom, left top, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, 10%, cover, auto, auto, auto;
}

tc {
  background-clip: content-box, border-box, padding-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, padding-box;
  background-position: left top, 5px bottom, 5px bottom, left top;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

uc {
  background-clip: content-box, padding-box, padding-box, border-box, padding-box, content-box, border-box, border-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: content-box, padding-box, content-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box;
  background-position: center, 10%, left top, left top, 10%, left top, center, 10%, left top, 5px bottom, 5px bottom, left top, 10%, left top, 10%, center, center;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, cover, auto, cover, auto, auto, auto, auto;
}

vc {
  background-clip: border-box, border-box, border-box, border-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, none /* FIXME */, none, none, none, none;
  background-origin: padding-box, padding-box, padding-box, border-box, content-box, padding-box;
  background-position: 10%, 5px bottom, left top, left top, left top, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

wc {
  background-clip: padding-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, border-box;
  background-position: left top, left top, 5px bottom;
  background-repeat: repeat, repeat, round no-repeat;
  background-size: auto, auto, auto;
}

xc {
  background-clip: border-box;
  background-color: initial;
  background-image: none /* FIXME */;
  background-origin: border-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

yc {
  background-clip: border-box;
  background-color: @home;
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

zc {
  background-clip: border-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

ad {
  background-clip: border-box, padding-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: border-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, left top, 10%, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat;
  background-size: cover, auto, auto, auto, auto;
}

bd {
  background-clip: border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: 10%;
  background-repeat: round no-repeat;
  background-size: auto;
}

cd {
  background-clip: border-box, padding-box, padding-box, border-box, padding-box, content-box, content-box, content-box, border-box, border-box, border-box, border-box, border-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, border-box, padding-box, content-box, padding-box, content-box, padding-box, border-box, border-box, padding-box, padding-box;
  background-position: left top, center, 5px bottom, left top, left top, left top, left top, left top, center, left top, left top, 10%, left top;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto;
}

dd {
  background-clip: content-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: border-box, border-box;
  background-position: left top, left top;
  background-repeat: repeat, round no-repeat;
  background-size: auto, auto;
}

ed {
  background-clip: content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: content-box;
  background-position: center;
  background-repeat: repeat;
  background-size: auto;
}

fd {
  background-clip: border-box, border-box, border-box, border-box, padding-box, content-box, border-box, border-box, content-box, padding-box, border-box, border-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box, content-box, padding-box, padding-box, padding-box, content-box;
  background-position: left top, 10%, 5px bottom, left top, left top, center, left top, left top, left top, center, 10%, left top, center;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, 10%;
}

gd {
  background-clip: border-box, border-box, padding-box, border-box, content-box, content-box;
  background-color: initial;
  background-image: none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, padding-box, padding-box, content-box, content-box;
  background-position: 10%, 10%, 5px bottom, 10%, 5px bottom, 10%;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: 10%, auto, auto, auto, auto, cover;
}

hd {
  background-clip: content-box, padding-box, padding-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, content-box;
  background-position: left top, 5px bottom, 5px bottom, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, cover, auto, auto;
}

id {
  background-clip: padding-box, border-box, padding-box, border-box, border-box, padding-box, border-box;
  background-color: @home;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, border-box, padding-box, border-box, border-box, border-box, content-box;
  background-position: 10%, 5px bottom, center, left top, left top, 10%, 10%;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: cover, auto, cover, auto, auto, auto, auto;
}

jd {
  background-clip: border-box, padding-box;
  background-color: currentColor;
  background-image: none, none;
  background-origin: border-box, content-box;
  background-position: 10%, left top;
  background-repeat: repeat, round no-repeat;
  background-size: auto, auto;
}

kd {
  background-clip: padding-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: padding-box;
  background-position: center;
  background-repeat: repeat;
  background-size: 10%;
}

ld {
  background-clip: border-box, border-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: border-box, padding-box, content-box;
  background-position: left top, 5px bottom, 5px bottom;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, cover;
}

md {
  background-clip: border-box, padding-box, content-box, border-box, padding-box, border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, content-box, border-box, padding-box, border-box, border-box;
  background-position: left top, left top, 10%, left top, center, left top, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

nd {
  background-clip: border-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box;
  background-position: 5px bottom, 5px bottom;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

od {
  background-clip: border-box, padding-box, border-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: center, left top, left top, 5px bottom, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto;
}

pd {
  background-clip: content-box;
  background-color: @home;
  background-image: none;
  background-origin: content-box;
  background-position: 10%;
  background-repeat: round no-repeat;
  background-size: auto;
}

qd {
  background-clip: border-box, padding-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none;
  background-origin: padding-box, content-box, padding-box;
  background-position: 5px bottom, center, center;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: auto, auto, 10%;
}

rd {
  background-clip: border-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, none;
  background-origin: border-box, content-box;
  background-position: left top, 10%;
  background-repeat: repeat, round no-repeat;
  background-size: auto, cover;
}

sd {
  background-clip: border-box, content-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box, border-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, border-box, padding-box, padding-box, padding-box, content-box, padding-box, border-box, padding-box, padding-box;
  background-position: 10%, 10%, left top, 5px bottom, left top, 10%, left top, 5px bottom, center, center, 5px bottom;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, 10%, auto, auto, 10%, 10%, auto, 10%;
}

td {
  background-clip: padding-box, border-box, border-box, border-box, border-box, padding-box, border-box, border-box, border-box, padding-box, border-box, border-box, padding-box, padding-box, padding-box, border-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none, none, none /* FIXME */, none, none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, border-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 10%, left top, left top, left top, center, left top, 10%, 5px bottom, left top, center, left top, left top, left top, left top, left top, 10%, 10%;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto;
}

ud {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: border-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: 10%;
}

vd {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

wd {
  background-clip: border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box;
  background-position: 10%, 5px bottom;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

xd {
  background-clip: border-box, padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, padding-box;
  background-position: left top, left top;
  background-repeat: repeat, round no-repeat;
  background-size: auto, auto;
}

yd {
  background-clip: content-box, padding-box, border-box, padding-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, padding-box, border-box, padding-box, content-box;
  background-position: left top, 10%, 10%, 10%, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, 10%, 10%;
}

zd {
  background-clip: content-box, content-box;
  background-color: initial;
  background-image: none, none;
  background-origin: content-box, content-box;
  background-position: 10%, 5px bottom;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

ae {
  background-clip: content-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box, border-box, content-box, padding-box, border-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */, none;
  background-origin: content-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box, content-box, content-box, padding-box, padding-box, padding-box;
  background-position: 10%, 5px bottom, center, left top, left top, 5px bottom, 10%, center, left top, left top, center, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, cover, auto, auto, auto, auto, cover, auto, auto, auto, 10%, auto;
}

be {
  background-clip: border-box, border-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, padding-box;
  background-position: 5px bottom, 5px bottom;
  background-repeat: repeat, round no-repeat;
  background-size: auto, auto;
}

ce {
  background-clip: padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

de {
  background-clip: border-box, border-box, border-box, padding-box, border-box, padding-box, border-box, border-box, border-box, padding-box, border-box, padding-box, padding-box, border-box, content-box;
  background-color: @home;
  background-image: none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: border-box, padding-box, content-box, padding-box, border-box, content-box, padding-box, padding-box, border-box, border-box, border-box, border-box, content-box, border-box, content-box;
  background-position: 10%, center, center, left top, center, 5px bottom, left top, left top, left top, left top, 5px bottom, center, left top, 10%, 5px bottom;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

ee {
  background-clip: border-box, border-box, padding-box, content-box, padding-box, border-box, border-box, content-box, padding-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, padding-box, content-box, padding-box, padding-box, border-box, content-box, padding-box, padding-box;
  background-position: 10%, left top, left top, left top, 5px bottom, 10%, center, left top, center, center;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

fe {
  background-clip: padding-box, padding-box, border-box, border-box, border-box, content-box, content-box, padding-box, content-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: border-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, content-box;
  background-position: left top, left top, 5px bottom, center, 10%, left top, center, center, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, cover, 10%;
}

ge {
  background-clip: padding-box, border-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, content-box;
  background-position: 10%, left top, 10%;
  background-repeat: round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, 10%;
}

he {
  background-clip: content-box, border-box, border-box, content-box, content-box, content-box, content-box, padding-box, border-box;
  background-color: @home;
  background-image: none, none, none, none, none, none, none, none, none;
  background-origin: content-box, border-box, padding-box, content-box, content-box, padding-box, content-box, padding-box, border-box;
  background-position: left top, center, 10%, 5px bottom, left top, left top, 10%, center, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, 10%;
}

ie {
  background-clip: padding-box, border-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, border-box, border-box, padding-box, padding-box;
  background-position: 10%, center, center, left top, 5px bottom;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: cover, 10%, auto, auto, cover;
}

je {
  background-clip: content-box, border-box, border-box, border-box, content-box, border-box, border-box, border-box, border-box, content-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, padding-box, content-box, padding-box, border-box, border-box, padding-box, content-box, border-box, padding-box;
  background-position: center, left top, left top, 10%, 10%, center, left top, center, left top, 5px bottom, left top, 5px bottom;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, cover, auto, auto, auto, cover, auto, auto, auto, cover;
}

ke {
  background-clip: padding-box, content-box, content-box;
  background-color: @home;
  background-image: none /* FIXME */, none, none;
  background-origin: padding-box, border-box, content-box;
  background-position: 5px bottom, left top, left top;
  background-repeat: round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto;
}

le {
  background-clip: border-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: cover;
}

me {
  background-clip: border-box, border-box, padding-box, border-box, padding-box, content-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none, none, none, none;
  background-origin: content-box, border-box, padding-box, border-box, border-box, border-box;
  background-position: 5px bottom, left top, left top, center, left top, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: cover, auto, auto, auto, auto, auto;
}

ne {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: border-box;
  background-position: 10%;
  background-repeat: round no-repeat;
  background-size: 10%;
}

oe {
  background-clip: border-box, padding-box, padding-box, content-box, border-box, content-box, border-box, content-box, border-box, content-box, padding-box, padding-box, padding-box;
  background-color: currentColor;
  background-image: none, none, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none;
  background-origin: border-box, padding-box, padding-box, content-box, border-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, 5px bottom, center, left top, left top, 10%, left top, left top, left top, 5px bottom, left top, 5px bottom, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, cover, auto, auto, auto, auto, auto, auto, auto;
}

pe {
  background-clip: border-box, content-box, content-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none /* FIXME */, none, none, none, none;
  background-origin: padding-box, content-box, padding-box, padding-box, padding-box, padding-box;
  background-position: left top, left top, 5px bottom, center, left top, 5px bottom;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

qe {
  background-clip: border-box;
  background-color: initial;
  background-image: none /* FIXME */;
  background-origin: border-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

re {
  background-clip: border-box, border-box, border-box, border-box, padding-box, padding-box, border-box, border-box, padding-box, border-box, padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box, border-box, border-box, content-box;
  background-position: left top, 5px bottom, left top, left top, left top, left top, 10%, 5px bottom, center, 10%, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: cover, auto, auto, auto, auto, auto, 10%, auto, auto, auto, 10%;
}

se {
  background-clip: content-box, border-box, border-box, padding-box, border-box, border-box, border-box, content-box, padding-box, border-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, border-box, content-box, border-box, content-box, padding-box, padding-box, padding-box, padding-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none /* FIXME */, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, border-box, border-box, padding-box, border-box, padding-box, border-box, content-box, padding-box, border-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, border-box, content-box, padding-box, content-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: center, center, center, left top, 5px bottom, 5px bottom, 10%, center, center, left top, 5px bottom, center, left top, center, left top, 5px bottom, 10%, 5px bottom, 5px bottom, 5px bottom, 5px bottom, 5px bottom, 10%, center, left top, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, cover, cover, auto, auto, auto, auto;
}

te {
  background-clip: border-box, border-box, border-box, padding-box, content-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box;
  background-color: currentColor;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, padding-box, content-box, padding-box, content-box, padding-box, padding-box, padding-box, content-box, content-box;
  background-position: left top, left top, left top, center, 5px bottom, center, left top, 10%, center, 10%, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, cover, auto, auto, auto, cover, auto, auto, auto, auto, auto;
}

ue {
  background-clip: content-box;
  background-color: @home;
  background-image: none;
  background-origin: padding-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: auto;
}

ve {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

we {
  background-clip: padding-box;
  background-color: currentColor;
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

xe {
  background-clip: content-box, padding-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: content-box, padding-box, padding-box;
  background-position: center, left top, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: 10%, auto, auto;
}

ye {
  background-clip: content-box, content-box, content-box, padding-box, content-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, content-box, content-box, padding-box, content-box;
  background-position: left top, left top, left top, left top, left top;
  background-repeat: repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto;
}

ze {
  background-clip: border-box, border-box, content-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, border-box, content-box, padding-box;
  background-position: 10%, center, 5px bottom, 10%;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

af {
  background-clip: border-box, padding-box, border-box, content-box, border-box, border-box, content-box, padding-box, border-box, padding-box, content-box, content-box, border-box, content-box, border-box, padding-box, border-box, content-box, border-box, border-box, border-box, padding-box, border-box, content-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none /* FIXME */, none;
  background-origin: padding-box, content-box, padding-box, content-box, padding-box, border-box, content-box, padding-box, border-box, padding-box, content-box, content-box, border-box, content-box, border-box, padding-box, padding-box, content-box, padding-box, border-box, border-box, padding-box, padding-box, content-box, padding-box;
  background-position: left top, left top, 10%, left top, left top, center, left top, left top, left top, 10%, left top, center, center, center, left top, left top, left top, left top, left top, left top, left top, left top, left top, left top, center;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, 10%, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto;
}

bf {
  background-clip: padding-box;
  background-color: initial;
  background-image: none /* FIXME */;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: 10%;
}

cf {
  background-clip: border-box, content-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: padding-box, content-box, border-box, border-box;
  background-position: left top, left top, 5px bottom, left top;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, auto;
}

df {
  background-clip: border-box, border-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, border-box, padding-box;
  background-position: left top, 10%, 10%;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

ef {
  background-clip: border-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

ff {
  background-clip: border-box, border-box, border-box, content-box, padding-box, border-box, content-box, padding-box, border-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, none /* FIXME */, none, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, content-box, border-box, padding-box, border-box, padding-box, padding-box;
  background-position: 5px bottom, left top, center, left top, center, left top, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, cover, auto;
}

gf {
  background-clip: padding-box, content-box, border-box, border-box;
  background-color: @home;
  background-image: none /* FIXME */, none /* FIXME */, none /* FIXME */, none;
  background-origin: padding-box, padding-box, border-box, border-box;
  background-position: 10%, 10%, left top, 10%;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, cover, 10%, auto;
}

hf {
  background-clip: padding-box, border-box, padding-box, content-box, border-box, padding-box, padding-box, padding-box;
  background-color: currentColor;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, content-box, border-box, padding-box, padding-box, content-box, padding-box;
  background-position: center, 5px bottom, left top, left top, center, 5px bottom, left top, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: cover, auto, auto, auto, 10%, auto, auto, auto;
}

if {
  background-clip: padding-box, content-box, content-box, border-box, border-box, padding-box, padding-box, content-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none;
  background-origin: padding-box, content-box, content-box, border-box, padding-box, border-box, padding-box, border-box, padding-box, border-box, padding-box;
  background-position: left top, center, left top, 10%, 5px bottom, 5px bottom, left top, center, center, left top, 10%;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, 10%, auto, auto, 10%, cover, auto, auto, auto, auto, auto;
}

jf {
  background-clip: content-box, content-box, border-box, content-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none /* FIXME */, none, none /* FIXME */;
  background-origin: content-box, content-box, border-box, content-box, border-box, padding-box;
  background-position: center, left top, 5px bottom, 5px bottom, 10%, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, cover, cover, auto, auto;
}

kf {
  background-clip: border-box, padding-box, border-box, padding-box, border-box, content-box;
  background-color: @home;
  background-image: none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, border-box, padding-box, padding-box, padding-box, border-box;
  background-position: 5px bottom, left top, 5px bottom, 10%, 10%, 5px bottom;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, cover, 10%, auto;
}

lf {
  background-clip: content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: content-box;
  background-position: center;
  background-repeat: repeat;
  background-size: auto;
}

mf {
  background-clip: padding-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: auto;
}

nf {
  background-clip: border-box;
  background-color: currentColor;
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

of {
  background-clip: content-box, border-box, border-box, content-box, border-box, padding-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none;
  background-origin: content-box, border-box, padding-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box;
  background-position: left top, left top, 10%, center, center, left top, 5px bottom, center, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, 10%, cover, auto, auto, cover, cover, auto, auto;
}

pf {
  background-clip: content-box, border-box, content-box, border-box, border-box, border-box, padding-box, content-box, content-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, content-box, padding-box, content-box, padding-box, padding-box, content-box, content-box, padding-box, padding-box;
  background-position: center, left top, 5px bottom, left top, left top, left top, 5px bottom, 5px bottom, 10%, 10%, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

qf {
  background-clip: border-box, border-box, border-box, content-box, border-box, content-box, border-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, border-box, padding-box, padding-box, content-box, padding-box, border-box;
  background-position: 10%, left top, 5px bottom, left top, left top, left top, center, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: cover, auto, auto, auto, auto, auto, auto, auto;
}

rf {
  background-clip: border-box, border-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, content-box;
  background-position: 5px bottom, left top, center;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, 10%, cover;
}

sf {
  background-clip: padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

tf {
  background-clip: padding-box, content-box, padding-box, border-box, border-box, content-box, content-box, padding-box, border-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none;
  background-origin: content-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box, border-box, border-box, border-box;
  background-position: center, 5px bottom, left top, left top, center, 10%, 10%, 5px bottom, 10%, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, cover, auto, auto, 10%, auto, auto, auto, auto;
}

uf {
  background-clip: border-box;
  background-color: currentColor;
  background-image: none /* FIXME */;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

vf {
  background-clip: content-box, border-box, border-box, padding-box, border-box, padding-box, border-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none /* FIXME */, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none;
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: center, left top, 10%, 10%, left top, left top, 10%, left top, center, center;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

wf {
  background-clip: content-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, none;
  background-origin: content-box, padding-box, padding-box;
  background-position: 10%, left top, center;
  background-repeat: repeat, repeat, repeat;
  background-size: cover, auto, auto;
}

xf {
  background-clip: border-box, content-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, border-box, padding-box, padding-box;
  background-position: left top, 10%, 5px bottom, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

yf {
  background-clip: content-box, border-box, padding-box, padding-box, padding-box, content-box, content-box, content-box, border-box, border-box, padding-box, border-box, border-box, border-box, content-box, content-box, padding-box, border-box, padding-box;
  background-color: @home;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, padding-box, padding-box, content-box, content-box, content-box, border-box, padding-box, border-box, padding-box, border-box, padding-box, border-box, padding-box, border-box, content-box, padding-box, padding-box;
  background-position: left top, center, center, left top, left top, left top, left top, center, 5px bottom, 5px bottom, left top, left top, left top, 10%, left top, left top, left top, left top, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

zf {
  background-clip: border-box, border-box;
  background-color: currentColor;
  background-image: none, none;
  background-origin: border-box, border-box;
  background-position: left top, left top;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

ag {
  background-clip: border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

bg {
  background-clip: content-box, border-box, padding-box, border-box, border-box, border-box, padding-box, border-box, content-box, border-box;
  background-color: currentColor;
  background-image: none, none, none /* FIXME */, none, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, padding-box, border-box, border-box, padding-box, border-box, padding-box, border-box, content-box, border-box;
  background-position: 10%, left top, left top, 10%, left top, left top, 10%, left top, 10%, left top;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

cg {
  background-clip: border-box, border-box, padding-box, border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, none, none;
  background-origin: border-box, padding-box, content-box, content-box, content-box;
  background-position: center, 10%, 5px bottom, 10%, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, cover, auto, 10%, 10%;
}

dg {
  background-clip: padding-box, content-box;
  background-color: initial;
  background-image: none, none;
  background-origin: padding-box, content-box;
  background-position: center, 10%;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

eg {
  background-clip: content-box, padding-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, border-box, padding-box;
  background-position: left top, center, left top;
  background-repeat: round no-repeat, round no-repeat, round no-repeat;
  background-size: 10%, auto, auto;
}

fg {
  background-clip: padding-box, border-box, padding-box;
  background-color: @home;
  background-image: none, none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, border-box, padding-box;
  background-position: center, 5px bottom, 10%;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

gg {
  background-clip: content-box, content-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, content-box;
  background-position: left top, 10%;
  background-repeat: round no-repeat, round no-repeat;
  background-size: auto, cover;
}

hg {
  background-clip: border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: 10%;
}

ig {
  background-clip: border-box, content-box, content-box, content-box, border-box, content-box, content-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none, none;
  background-origin: padding-box, border-box, content-box, content-box, padding-box, content-box, content-box, border-box, padding-box, padding-box;
  background-position: 10%, 10%, 10%, left top, left top, center, left top, 10%, 5px bottom, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto;
}

jg {
  background-clip: border-box, padding-box, padding-box, border-box, border-box, content-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: border-box, padding-box, padding-box, border-box, padding-box, content-box;
  background-position: 10%, center, center, center, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, cover;
}

kg {
  background-clip: padding-box, border-box, padding-box, border-box, border-box, content-box, padding-box, border-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box;
  background-position: left top, left top, center, left top, 10%, 10%, 10%, left top;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: cover, auto, auto, auto, auto, auto, auto, auto;
}

lg {
  background-clip: border-box, content-box, content-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, content-box;
  background-position: 5px bottom, left top, 10%;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, 10%;
}

mg {
  background-clip: border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

ng {
  background-clip: border-box, content-box, border-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, padding-box, padding-box, padding-box, padding-box;
  background-position: center, left top, 10%, left top, center, 10%;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: cover, auto, auto, auto, auto, cover;
}

og {
  background-clip: border-box, content-box, content-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, content-box, content-box, padding-box;
  background-position: 10%, center, center, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

pg {
  background-clip: border-box, padding-box, content-box, border-box, border-box, padding-box, border-box, border-box, border-box, content-box, border-box, padding-box, content-box, content-box, border-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, border-box, border-box, border-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box, content-box, content-box, padding-box;
  background-position: 5px bottom, 10%, left top, 5px bottom, center, left top, center, left top, 10%, 5px bottom, 10%, 5px bottom, left top, left top, center;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, auto, auto, 10%, 10%, auto, auto, auto, cover, auto, auto, auto, auto;
}

qg {
  background-clip: content-box, content-box, border-box, padding-box, border-box, padding-box, content-box, padding-box, content-box, border-box, padding-box, border-box, border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */, none /* FIXME */, none, none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, padding-box, border-box, padding-box, content-box, padding-box, padding-box, border-box, padding-box, border-box, border-box, padding-box;
  background-position: left top, left top, 5px bottom, left top, center, left top, center, 10%, center, left top, 5px bottom, 5px bottom, left top, left top;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, cover, auto, auto, auto, auto, cover, auto, auto, cover, auto, auto, auto, auto;
}

rg {
  background-clip: padding-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: content-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

sg {
  background-clip: border-box, padding-box, content-box, content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, none, none, none;
  background-origin: border-box, content-box, border-box, content-box, padding-box;
  background-position: left top, 5px bottom, 10%, left top, left top;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, cover, auto, auto, auto;
}

tg {
  background-clip: content-box, border-box, content-box;
  background-color: currentColor;
  background-image: none, none, none;
  background-origin: border-box, padding-box, content-box;
  background-position: left top, left top, center;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

ug {
  background-clip: border-box, content-box, border-box, content-box, border-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, padding-box, content-box, border-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box;
  background-position: 10%, 10%, 5px bottom, center, center, left top, left top, left top, 5px bottom, 10%, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, 10%, auto, auto, auto, auto, 10%, auto, cover;
}

vg {
  background-clip: border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, none;
  background-origin: padding-box, padding-box;
  background-position: 5px bottom, left top;
  background-repeat: round no-repeat, round no-repeat;
  background-size: auto, cover;
}

wg {
  background-clip: padding-box, border-box, padding-box, content-box, border-box, border-box, border-box, border-box, border-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, padding-box, content-box, content-box, padding-box, padding-box, content-box, padding-box, padding-box, content-box, content-box;
  background-position: 10%, left top, center, 5px bottom, left top, 5px bottom, 10%, left top, 10%, 5px bottom, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: cover, auto, auto, auto, auto, 10%, 10%, auto, cover, auto, auto, auto;
}

xg {
  background-clip: border-box, border-box, padding-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, 5px bottom, 5px bottom, left top, 5px bottom, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, cover, auto, auto, auto, auto;
}

yg {
  background-clip: border-box, border-box, padding-box, content-box, border-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, border-box, padding-box, padding-box, padding-box, content-box, padding-box, border-box;
  background-position: center, 10%, 5px bottom, left top, left top, left top, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto;
}

zg {
  background-clip: content-box, content-box, content-box, content-box, border-box, padding-box, content-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none, none /* FIXME */, none, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none;
  background-origin: content-box, content-box, border-box, content-box, padding-box, padding-box, content-box, padding-box, border-box;
  background-position: 10%, left top, center, 5px bottom, center, left top, left top, center, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: 10%, cover, cover, auto, auto, auto, auto, auto, auto;
}

ah {
  background-clip: padding-box, content-box, border-box, padding-box, padding-box, padding-box, border-box, content-box, border-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */;
  background-origin: padding-box, border-box, border-box, padding-box, content-box, content-box, padding-box, content-box, padding-box;
  background-position: 5px bottom, center, center, 10%, left top, 10%, left top, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, cover, auto, auto, auto, auto, auto;
}

bh {
  background-clip: padding-box, border-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, border-box;
  background-position: center, left top, left top;
  background-repeat: repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, cover;
}

ch {
  background-clip: border-box, border-box, border-box, content-box, content-box, content-box, content-box, border-box, border-box, padding-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none;
  background-origin: padding-box, padding-box, padding-box, content-box, padding-box, content-box, content-box, border-box, border-box, padding-box;
  background-position: left top, 5px bottom, 10%, left top, left top, center, 5px bottom, 5px bottom, 10%, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, cover, auto, auto, auto, auto, auto, cover, auto;
}

dh {
  background-clip: padding-box, content-box, padding-box, padding-box, padding-box;
  background-color: @home;
  background-image: none, none, none /* FIXME */, none, none;
  background-origin: padding-box, content-box, border-box, padding-box, content-box;
  background-position: 5px bottom, center, 10%, 10%, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat;
  background-size: 10%, 10%, auto, auto, auto;
}

eh {
  background-clip: border-box, padding-box, padding-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, none /* FIXME */, none, none;
  background-origin: border-box, border-box, content-box, padding-box;
  background-position: 5px bottom, left top, left top, left top;
  background-repeat: repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, cover, auto;
}

fh {
  background-clip: padding-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none /* FIXME */;
  background-origin: padding-box, padding-box;
  background-position: center, 5px bottom;
  background-repeat: round no-repeat, round no-repeat;
  background-size: auto, auto;
}

gh {
  background-clip: border-box, border-box;
  background-color: initial;
  background-image: none, none;
  background-origin: padding-box, border-box;
  background-position: center, 5px bottom;
  background-repeat: repeat, round no-repeat;
  background-size: auto, auto;
}

hh {
  background-clip: border-box, content-box, border-box, content-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, none, none, none;
  background-origin: padding-box, padding-box, border-box, content-box, padding-box;
  background-position: left top, left top, center, 5px bottom, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto;
}

ih {
  background-clip: border-box, border-box;
  background-color: initial;
  background-image: none, none;
  background-origin: padding-box, border-box;
  background-position: left top, left top;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

jh {
  background-clip: content-box, padding-box, content-box, border-box, padding-box, border-box, content-box, content-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, content-box;
  background-position: center, 5px bottom, center, 10%, center, left top, center, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, auto, cover, auto, 10%;
}

kh {
  background-clip: padding-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

lh {
  background-clip: border-box, border-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, border-box;
  background-position: left top, center;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

mh {
  background-clip: border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box;
  background-position: 10%;
  background-repeat: round no-repeat;
  background-size: auto;
}

nh {
  background-clip: border-box, border-box, border-box, border-box, padding-box, border-box, border-box, padding-box, content-box, padding-box;
  background-color: currentColor;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, padding-box, content-box, padding-box, padding-box, border-box, padding-box, content-box, border-box;
  background-position: center, left top, center, center, 5px bottom, 5px bottom, 10%, left top, 5px bottom, center;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, 10%, cover, auto, auto, auto, auto;
}

oh {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: border-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: auto;
}

ph {
  background-clip: padding-box, content-box, padding-box, border-box, border-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, none /* FIXME */, none, none /* FIXME */, none, none /* FIXME */;
  background-origin: padding-box, content-box, padding-box, padding-box, padding-box, padding-box;
  background-position: center, left top, 10%, 10%, 5px bottom, center;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat;
  background-size: 10%, auto, cover, auto, cover, 10%;
}

qh {
  background-clip: content-box, border-box, border-box, content-box, border-box, border-box, padding-box, border-box, border-box, padding-box, content-box, padding-box, padding-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none /* FIXME */, none, none, none, none /* FIXME */, none /* FIXME */, none, none, none, none, none;
  background-origin: content-box, border-box, padding-box, border-box, padding-box, border-box, border-box, padding-box, border-box, padding-box, content-box, padding-box, padding-box, border-box, content-box, border-box;
  background-position: left top, center, left top, left top, left top, center, center, 5px bottom, 10%, 10%, left top, 10%, 5px bottom, left top, 10%, 10%;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, 10%, auto, auto, auto, auto;
}

rh {
  background-clip: border-box, border-box, content-box;
  background-color: @home;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, content-box;
  background-position: 10%, left top, left top;
  background-repeat: round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto;
}

sh {
  background-clip: padding-box, border-box, border-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: padding-box, border-box, padding-box, padding-box;
  background-position: center, left top, 10%, 10%;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, cover;
}

th {
  background-clip: content-box, content-box, border-box, padding-box, border-box, content-box, border-box;
  background-color: currentColor;
  background-image: none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, content-box, border-box, border-box, border-box, content-box, padding-box;
  background-position: left top, 5px bottom, left top, left top, left top, center, 10%;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, cover, auto, auto, auto, auto, auto;
}

uh {
  background-clip: padding-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: round no-repeat;
  background-size: auto;
}

vh {
  background-clip: padding-box, content-box, padding-box, content-box, padding-box, padding-box, border-box, border-box, border-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, content-box, content-box, content-box, padding-box, padding-box, border-box, content-box, padding-box, border-box;
  background-position: left top, 5px bottom, 5px bottom, 10%, left top, left top, 5px bottom, left top, 5px bottom, 5px bottom;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto;
}

wh {
  background-clip: content-box, content-box, content-box, border-box, content-box, padding-box, border-box, content-box, content-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: content-box, content-box, content-box, padding-box, content-box, padding-box, padding-box, content-box, content-box, padding-box, padding-box, padding-box;
  background-position: 10%, left top, 5px bottom, 10%, left top, left top, left top, 5px bottom, center, center, 5px bottom, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

xh {
  background-clip: border-box, border-box, border-box, padding-box, padding-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none;
  background-origin: border-box, content-box, padding-box, border-box, padding-box, padding-box;
  background-position: center, 5px bottom, left top, left top, 10%, center;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, 10%;
}

yh {
  background-clip: border-box, border-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box;
  background-position: left top, 10%;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

zh {
  background-clip: padding-box, padding-box, padding-box, border-box, padding-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */, none;
  background-origin: padding-box, content-box, padding-box, border-box, border-box, padding-box, border-box;
  background-position: 10%, center, center, 10%, 10%, center, 10%;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, 10%, auto, cover, auto;
}

ai {
  background-clip: padding-box, padding-box, content-box, border-box, border-box, content-box, border-box, border-box, padding-box, padding-box, padding-box, border-box, border-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none, none, none /* FIXME */, none;
  background-origin: padding-box, padding-box, content-box, border-box, border-box, content-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box;
  background-position: 10%, center, 10%, 5px bottom, left top, left top, left top, 5px bottom, center, 5px bottom, 10%, left top, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, 10%, cover, auto, 10%, 10%, cover, cover, auto, auto, cover, auto;
}

bi {
  background-clip: content-box, padding-box, padding-box, content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, content-box;
  background-position: center, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: cover, auto, auto, auto;
}

ci {
  background-clip: border-box, border-box, border-box, border-box, border-box, border-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, border-box, border-box, padding-box, padding-box, padding-box, border-box, padding-box;
  background-position: center, 10%, 10%, center, left top, left top, left top, center;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: cover, auto, auto, auto, auto, auto, auto, auto;
}

di {
  background-clip: border-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */;
  background-origin: padding-box, padding-box;
  background-position: left top, 10%;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

ei {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: border-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

fi {
  background-clip: border-box;
  background-color: @home;
  background-image: none;
  background-origin: content-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

gi {
  background-clip: padding-box, content-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none;
  background-origin: border-box, padding-box, border-box, padding-box;
  background-position: 5px bottom, center, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: cover, auto, auto, auto;
}

hi {
  background-clip: content-box, content-box;
  background-color: @home;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box;
  background-position: 10%, left top;
  background-repeat: repeat, repeat;
  background-size: cover, auto;
}

ii {
  background-clip: content-box, padding-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, padding-box;
  background-position: left top, 5px bottom;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

ji {
  background-clip: padding-box, content-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none /* FIXME */;
  background-origin: padding-box, content-box, border-box, padding-box;
  background-position: 5px bottom, left top, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto;
}

ki {
  background-clip: border-box, border-box, border-box, content-box, border-box, content-box, border-box, border-box, padding-box, padding-box, border-box, content-box, border-box;
  background-color: @home;
  background-image: none, none /* FIXME */, none, none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, border-box, content-box, border-box, content-box, content-box, padding-box, content-box, padding-box, padding-box, content-box, border-box;
  background-position: 5px bottom, 10%, left top, left top, 5px bottom, left top, left top, 5px bottom, 10%, left top, center, left top, center;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, 10%, auto, 10%, auto, cover, auto, auto, auto;
}

li {
  background-clip: border-box, padding-box, content-box, content-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, none, none, none /* FIXME */;
  background-origin: border-box, padding-box, content-box, content-box, padding-box;
  background-position: left top, 10%, left top, 5px bottom, center;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto;
}

mi {
  background-clip: padding-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, border-box, padding-box, border-box;
  background-position: left top, 10%, 10%, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

ni {
  background-clip: border-box, border-box, padding-box, border-box, padding-box, padding-box, border-box, content-box, border-box, padding-box, border-box, padding-box, border-box, padding-box, border-box, padding-box, content-box, border-box, border-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, content-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, border-box, border-box, padding-box, content-box, padding-box, padding-box;
  background-position: left top, center, 5px bottom, left top, center, left top, center, left top, left top, left top, 10%, left top, left top, center, left top, 5px bottom, left top, center, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, 10%, 10%, cover, auto, auto, auto, auto, 10%, auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto;
}

oi {
  background-clip: padding-box, content-box, border-box, padding-box, content-box, border-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, none, none, none /* FIXME */;
  background-origin: padding-box, content-box, padding-box, padding-box, content-box, border-box;
  background-position: 5px bottom, center, 10%, 10%, 10%, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, 10%, auto, auto;
}

pi {
  background-clip: content-box, content-box, border-box;
  background-color: @home;
  background-image: none, none, none;
  background-origin: content-box, content-box, border-box;
  background-position: 5px bottom, left top, left top;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

qi {
  background-clip: border-box, content-box, content-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, content-box, border-box;
  background-position: left top, 10%, 5px bottom;
  background-repeat: repeat, repeat, repeat;
  background-size: 10%, cover, auto;
}

ri {
  background-clip: content-box;
  background-color: initial;
  background-image: none;
  background-origin: content-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

si {
  background-clip: border-box, border-box, border-box, border-box, padding-box, padding-box;
  background-color: @home;
  background-image: none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, padding-box, border-box, padding-box, padding-box;
  background-position: left top, 5px bottom, left top, left top, 10%, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, 10%, auto, auto, 10%, auto;
}

ti {
  background-clip: padding-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: cover;
}

ui {
  background-clip: border-box, border-box, content-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none;
  background-origin: border-box, content-box, content-box, content-box;
  background-position: left top, 5px bottom, 10%, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

vi {
  background-clip: border-box, content-box;
  background-color: currentColor;
  background-image: none, none;
  background-origin: padding-box, content-box;
  background-position: center, 5px bottom;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

wi {
  background-clip: border-box, border-box, border-box, padding-box, border-box, padding-box, padding-box, content-box, border-box, content-box, border-box, border-box, border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none;
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, border-box, content-box, padding-box, border-box, border-box, padding-box;
  background-position: center, left top, center, left top, 5px bottom, left top, left top, left top, 10%, 5px bottom, 10%, 10%, 10%, center;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, cover, 10%, auto, cover;
}

xi {
  background-clip: padding-box, border-box, border-box, padding-box, content-box, border-box, border-box, padding-box, padding-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none, none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box, border-box, padding-box, border-box, padding-box, padding-box, padding-box, border-box, padding-box;
  background-position: center, left top, left top, left top, left top, 10%, center, left top, center, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: cover, auto, auto, auto, auto, auto, auto, auto, cover, auto;
}

yi {
  background-clip: border-box, border-box, padding-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, padding-box, border-box;
  background-position: left top, 5px bottom, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto;
}

zi {
  background-clip: padding-box, padding-box, padding-box, content-box, content-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */;
  background-origin: border-box, padding-box, padding-box, content-box, content-box, padding-box;
  background-position: 5px bottom, left top, left top, left top, left top, 10%;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

aj {
  background-clip: padding-box, content-box, content-box, border-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none, none, none;
  background-origin: padding-box, content-box, content-box, padding-box;
  background-position: 5px bottom, left top, left top, center;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto;
}

bj {
  background-clip: content-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box;
  background-position: 10%, 5px bottom;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

cj {
  background-clip: border-box, border-box, border-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none /* FIXME */, none, none, none;
  background-origin: padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: 10%, 10%, left top, left top, left top;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, cover;
}

dj {
  background-clip: content-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box, padding-box, border-box, border-box, border-box, border-box, content-box, padding-box, border-box, content-box, content-box, border-box, padding-box, padding-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box, border-box, border-box, border-box, border-box, padding-box, border-box, border-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none;
  background-origin: content-box, padding-box, padding-box, padding-box, padding-box, content-box, content-box, content-box, border-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, content-box, content-box, border-box, padding-box, padding-box, border-box, border-box, border-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box, border-box, border-box, padding-box, padding-box, border-box, border-box;
  background-position: 5px bottom, left top, 10%, left top, 10%, left top, 5px bottom, 5px bottom, 5px bottom, 5px bottom, left top, left top, 5px bottom, left top, center, 10%, center, left top, center, 10%, 10%, 10%, left top, center, left top, 10%, 5px bottom, left top, 5px bottom, left top, center, left top, 10%, center, center, 10%;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto, auto, auto;
}

ej {
  background-clip: border-box, border-box, border-box, content-box, border-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none;
  background-origin: padding-box, border-box, padding-box, content-box, padding-box, padding-box, border-box;
  background-position: 5px bottom, center, 10%, center, left top, center, 10%;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, cover, 10%, auto, auto, auto;
}

fj {
  background-clip: content-box, padding-box, padding-box, content-box, content-box, content-box, content-box, padding-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none /* FIXME */, none, none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none;
  background-origin: content-box, padding-box, content-box, content-box, content-box, content-box, content-box, padding-box, padding-box;
  background-position: 10%, center, left top, 10%, left top, left top, center, center, center;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: cover, 10%, auto, auto, auto, auto, cover, auto, 10%;
}

gj {
  background-clip: border-box;
  background-color: currentColor;
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

hj {
  background-clip: border-box, content-box, content-box, padding-box, padding-box, border-box;
  background-color: initial;
  background-image: none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, content-box, content-box, padding-box, padding-box;
  background-position: 10%, center, center, left top, 5px bottom, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, cover;
}

ij {
  background-clip: padding-box, border-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none, none;
  background-origin: border-box, padding-box, padding-box, content-box;
  background-position: left top, left top, 10%, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, cover, auto;
}

jj {
  background-clip: content-box, padding-box, border-box, content-box, content-box, content-box, border-box, border-box, border-box, border-box, content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, padding-box, content-box, content-box, content-box, padding-box, border-box, padding-box, padding-box, content-box, padding-box;
  background-position: 10%, center, left top, left top, left top, left top, left top, left top, 5px bottom, center, left top, left top;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, cover;
}

kj {
  background-clip: border-box, content-box, border-box, border-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none /* FIXME */, none, none, none /* FIXME */, none, none, none, none;
  background-origin: padding-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box, content-box, content-box, padding-box, padding-box;
  background-position: 5px bottom, center, left top, left top, center, left top, left top, left top, center, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, 10%, auto, auto, auto, auto, auto, 10%, auto, auto;
}

lj {
  background-clip: content-box, padding-box, border-box, border-box, content-box, border-box, padding-box, border-box, content-box, content-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none, none, none, none, none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, content-box, content-box, border-box, padding-box, border-box, content-box, content-box, border-box;
  background-position: left top, left top, left top, 5px bottom, left top, 10%, 5px bottom, 5px bottom, left top, left top, 5px bottom;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, cover, auto, auto, auto, auto, 10%, auto, auto, auto;
}

mj {
  background-clip: border-box, border-box, content-box, content-box, padding-box, border-box, content-box, border-box, content-box, content-box, border-box, border-box, border-box, content-box, border-box, border-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, content-box, content-box, padding-box, padding-box, border-box, padding-box, content-box, content-box, border-box, padding-box, padding-box, content-box, border-box, padding-box;
  background-position: 5px bottom, center, center, 10%, left top, left top, center, left top, center, center, left top, left top, left top, 10%, 10%, left top;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, 10%, cover, auto, auto, auto, auto, 10%, cover, cover, auto, auto, auto, cover, auto, 10%;
}

nj {
  background-clip: padding-box, padding-box, padding-box, border-box, padding-box, content-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, content-box, border-box, padding-box, padding-box;
  background-position: center, 5px bottom, left top, 10%, left top, left top, 5px bottom, left top, left top;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, cover, auto, auto, auto, auto, 10%, auto, auto;
}

oj {
  background-clip: border-box, content-box, border-box, padding-box, border-box, padding-box, content-box, border-box;
  background-color: @home;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, center, center, left top, 5px bottom, left top, center, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: cover, 10%, auto, 10%, auto, auto, auto, auto;
}

pj {
  background-clip: border-box, content-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, content-box, border-box, border-box;
  background-position: center, left top, 10%, 5px bottom;
  background-repeat: repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, cover;
}

qj {
  background-clip: padding-box, content-box, content-box, padding-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none, none /* FIXME */;
  background-origin: padding-box, content-box, content-box, padding-box, content-box;
  background-position: left top, 10%, left top, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, auto;
}

rj {
  background-clip: padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: padding-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: auto;
}

sj {
  background-clip: content-box, padding-box, padding-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: none, none, none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: left top, 5px bottom, 10%, left top, 10%, 10%;
  background-repeat: repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, auto, auto;
}

tj {
  background-clip: border-box, content-box, content-box, padding-box, content-box, content-box, padding-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, content-box, padding-box, padding-box, content-box, content-box, border-box;
  background-position: center, 10%, 5px bottom, center, 10%, left top, 5px bottom;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

uj {
  background-clip: border-box, content-box, padding-box, padding-box, border-box, padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */;
  background-origin: padding-box, content-box, padding-box, padding-box, content-box, padding-box;
  background-position: 10%, 10%, 10%, 5px bottom, center, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, 10%, 10%, auto, auto;
}

vj {
  background-clip: border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: auto;
}

wj {
  background-clip: padding-box, border-box, border-box, border-box, content-box, border-box, border-box, border-box, border-box, border-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none, none /* FIXME */, none, none /* FIXME */;
  background-origin: padding-box, border-box, border-box, border-box, content-box, border-box, padding-box, border-box, border-box, border-box, border-box;
  background-position: 5px bottom, center, 5px bottom, center, 10%, center, left top, center, 10%, 10%, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, 10%, 10%, auto, auto, auto, 10%, auto, auto, auto;
}

xj {
  background-clip: padding-box, border-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box, padding-box, content-box, border-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, none, none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, border-box, padding-box, content-box, padding-box, content-box;
  background-position: center, 10%, left top, center, left top, 5px bottom, 5px bottom, left top, center, 5px bottom, left top, 5px bottom, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, cover, 10%, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto;
}

yj {
  background-clip: border-box;
  background-color: rgb(255,0,0);
  background-image: none;
  background-origin: border-box;
  background-position: 5px bottom;
  background-repeat: round no-repeat;
  background-size: 10%;
}

zj {
  background-clip: border-box, border-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, padding-box;
  background-position: 10%, center, left top;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

ak {
  background-clip: content-box, padding-box, padding-box, content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: content-box, padding-box, padding-box, border-box;
  background-position: 10%, left top, 5px bottom, 10%;
  background-repeat: repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, cover, auto;
}

bk {
  background-clip: border-box;
  background-color: currentColor;
  background-image: none;
  background-origin: border-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

ck {
  background-clip: content-box, content-box, border-box, padding-box, border-box, content-box, border-box, border-box, content-box, border-box, border-box, border-box, border-box, padding-box, border-box, padding-box, content-box, border-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, content-box, border-box, padding-box, border-box, content-box, border-box, content-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, content-box, padding-box;
  background-position: left top, 5px bottom, center, 10%, center, 5px bottom, left top, left top, left top, 5px bottom, 10%, left top, left top, left top, center, left top, 5px bottom, 10%, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, cover, auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto;
}

dk {
  background-clip: content-box;
  background-color: currentColor;
  background-image: none;
  background-origin: border-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: auto;
}

ek {
  background-clip: padding-box, border-box, border-box, border-box, border-box, border-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none, none, none, none /* FIXME */, none;
  background-origin: padding-box, padding-box, border-box, padding-box, padding-box, padding-box, content-box;
  background-position: 10%, 5px bottom, 5px bottom, left top, left top, center, 5px bottom;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: 10%, auto, auto, auto, auto, auto, auto;
}

fk {
  background-clip: padding-box, content-box, padding-box, border-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, content-box, border-box, padding-box, border-box, padding-box;
  background-position: left top, 5px bottom, center, 10%, 5px bottom, 10%;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

gk {
  background-clip: padding-box, padding-box, border-box, content-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none /* FIXME */, none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, border-box, border-box;
  background-position: 5px bottom, 10%, left top, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto;
}

hk {
  background-clip: content-box;
  background-color: currentColor;
  background-image: none;
  background-origin: content-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

ik {
  background-clip: content-box, padding-box, content-box, border-box, border-box, padding-box, border-box, content-box, content-box, border-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, padding-box, content-box, border-box, border-box, border-box, content-box, padding-box, padding-box;
  background-position: 10%, left top, left top, 10%, center, 10%, center, 5px bottom, left top, left top, 10%;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, cover, auto, auto, cover, auto, 10%, auto, auto, auto, cover;
}

jk {
  background-clip: content-box, content-box, border-box, border-box, content-box, border-box, padding-box, padding-box, border-box, border-box, padding-box, content-box, border-box;
  background-color: @home;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none;
  background-origin: content-box, content-box, padding-box, padding-box, content-box, padding-box, border-box, padding-box, padding-box, border-box, padding-box, content-box, padding-box;
  background-position: left top, center, center, left top, center, left top, 10%, left top, left top, left top, 10%, center, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

kk {
  background-clip: padding-box, border-box, border-box, padding-box, content-box, padding-box, padding-box, border-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none;
  background-origin: padding-box, content-box, border-box, padding-box, content-box, padding-box, padding-box, border-box, border-box;
  background-position: left top, center, left top, 10%, center, 10%, 10%, left top, 5px bottom;
  background-repeat: round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, 10%, auto, auto, auto, auto, auto;
}

lk {
  background-clip: content-box, content-box, content-box, border-box;
  background-color: initial;
  background-image: none, none, none, none;
  background-origin: content-box, content-box, padding-box, border-box;
  background-position: 10%, left top, 10%, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, 10%, auto;
}

mk {
  background-clip: padding-box, border-box, border-box, border-box, content-box, border-box, padding-box, content-box, content-box, border-box, border-box, border-box, border-box, padding-box, padding-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, padding-box, border-box, content-box, padding-box, padding-box, content-box, content-box, content-box, padding-box, border-box, border-box, padding-box, padding-box;
  background-position: center, left top, center, left top, left top, 5px bottom, 5px bottom, left top, 5px bottom, left top, 5px bottom, left top, 10%, 5px bottom, 10%;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, 10%, auto, 10%, auto, 10%, auto, cover, auto, auto, auto, auto, auto, auto;
}

nk {
  background-clip: border-box, content-box, padding-box, border-box, border-box, border-box, border-box, padding-box, padding-box, border-box, content-box, padding-box, padding-box, border-box, padding-box, content-box, border-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */;
  background-origin: border-box, content-box, padding-box, padding-box, border-box, border-box, padding-box, padding-box, padding-box, padding-box, content-box, content-box, padding-box, content-box, padding-box, content-box, padding-box;
  background-position: 10%, left top, left top, left top, left top, 5px bottom, left top, center, left top, 10%, 5px bottom, left top, 5px bottom, center, left top, left top, left top;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat;
  background-size: 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

ok {
  background-clip: content-box, padding-box, border-box, border-box, border-box, content-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none;
  background-origin: content-box, content-box, padding-box, border-box, padding-box, padding-box, padding-box, border-box, border-box;
  background-position: 10%, left top, center, 5px bottom, center, 5px bottom, 5px bottom, 5px bottom, 5px bottom;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: cover, auto, auto, 10%, auto, 10%, auto, auto, 10%;
}

pk {
  background-clip: padding-box, border-box, content-box, content-box, padding-box, padding-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, content-box, border-box, padding-box, content-box, padding-box;
  background-position: center, left top, left top, left top, 5px bottom, left top, center;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, 10%, auto, cover, auto, auto, auto;
}

qk {
  background-clip: border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

rk {
  background-clip: border-box, padding-box, border-box, padding-box, border-box, content-box, border-box, content-box, padding-box, border-box, border-box, border-box, padding-box, border-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none, none;
  background-origin: border-box, content-box, border-box, padding-box, padding-box, border-box, border-box, border-box, padding-box, border-box, padding-box, padding-box, padding-box, padding-box;
  background-position: center, center, 10%, left top, left top, left top, left top, center, left top, left top, 5px bottom, 5px bottom, 10%, center;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: cover, auto, 10%, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto, cover;
}

sk {
  background-clip: content-box, border-box, border-box, border-box, padding-box;
  background-color: currentColor;
  background-image: none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, border-box, padding-box, padding-box, padding-box;
  background-position: left top, center, 10%, 10%, center;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto;
}

tk {
  background-clip: border-box, content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none, none /* FIXME */;
  background-origin: padding-box, content-box, padding-box;
  background-position: 5px bottom, 10%, left top;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: 10%, auto, 10%;
}

uk {
  background-clip: border-box, padding-box, padding-box, border-box, border-box, border-box, padding-box, content-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, content-box, border-box, padding-box, padding-box, padding-box, border-box, content-box;
  background-position: 10%, center, center, center, left top, left top, left top, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, cover, auto, auto, auto, auto, auto, auto;
}

vk {
  background-clip: border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, none;
  background-origin: padding-box, padding-box;
  background-position: 5px bottom, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

wk {
  background-clip: content-box, padding-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, content-box, padding-box, padding-box, padding-box;
  background-position: left top, 10%, left top, center, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, 10%, auto;
}

xk {
  background-clip: border-box, padding-box;
  background-color: initial;
  background-image: none, none;
  background-origin: border-box, padding-box;
  background-position: left top, 5px bottom;
  background-repeat: repeat, repeat;
  background-size: auto, cover;
}

yk {
  background-clip: content-box;
  background-color: initial;
  background-image: none /* FIXME */;
  background-origin: border-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: auto;
}

zk {
  background-clip: border-box, border-box, padding-box, padding-box, content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, none /* FIXME */, none, none;
  background-origin: padding-box, border-box, padding-box, padding-box, content-box;
  background-position: 5px bottom, 5px bottom, 10%, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat;
  background-size: 10%, auto, auto, auto, auto;
}

al {
  background-clip: content-box, border-box, border-box, content-box;
  background-color: @home;
  background-image: none, none, none /* FIXME */, none /* FIXME */;
  background-origin: content-box, border-box, border-box, content-box;
  background-position: center, 5px bottom, 10%, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat;
  background-size: auto, 10%, cover, auto;
}

bl {
  background-clip: padding-box, border-box, padding-box, border-box, border-box, padding-box, padding-box, border-box, border-box, border-box, border-box, border-box, border-box, border-box, content-box, content-box, border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, none /* FIXME */, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, padding-box, border-box, padding-box, padding-box, content-box, border-box, border-box, padding-box, border-box, padding-box, border-box, border-box, content-box, content-box, padding-box, border-box;
  background-position: left top, 10%, 5px bottom, left top, 5px bottom, 5px bottom, 5px bottom, left top, 10%, 10%, 10%, left top, center, center, 5px bottom, 5px bottom, left top, 10%;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, cover, auto, auto, auto, auto;
}

cl {
  background-clip: border-box, border-box, padding-box, border-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none, none;
  background-origin: padding-box, padding-box, border-box, border-box, padding-box, border-box, border-box;
  background-position: left top, left top, 5px bottom, left top, left top, left top, center;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, cover;
}

dl {
  background-clip: padding-box, border-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box;
  background-position: 10%, 10%, left top;
  background-repeat: repeat, repeat, round no-repeat;
  background-size: auto, auto, auto;
}

el {
  background-clip: padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: round no-repeat;
  background-size: auto;
}

fl {
  background-clip: content-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none;
  background-origin: content-box, padding-box;
  background-position: 5px bottom, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

gl {
  background-clip: padding-box;
  background-color: currentColor;
  background-image: none;
  background-origin: border-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: cover;
}

hl {
  background-clip: border-box, padding-box, border-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none /* FIXME */, none, none /* FIXME */;
  background-origin: padding-box, padding-box, border-box, padding-box;
  background-position: 5px bottom, left top, 10%, center;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

il {
  background-clip: border-box, padding-box, content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, content-box, border-box;
  background-position: center, 5px bottom, center, 10%;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, cover;
}

jl {
  background-clip: padding-box, border-box, padding-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, padding-box;
  background-position: 5px bottom, left top, center;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: auto, auto, cover;
}

kl {
  background-clip: border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: cover;
}

ll {
  background-clip: border-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

ml {
  background-clip: content-box;
  background-color: initial;
  background-image: none /* FIXME */;
  background-origin: content-box;
  background-position: center;
  background-repeat: repeat;
  background-size: auto;
}

nl {
  background-clip: content-box, padding-box, border-box, border-box, content-box, border-box, content-box;
  background-color: initial;
  background-image: none, none, none, none, none /* FIXME */, none /* FIXME */, none;
  background-origin: padding-box, padding-box, border-box, padding-box, content-box, padding-box, content-box;
  background-position: left top, center, left top, left top, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

ol {
  background-clip: padding-box, padding-box, content-box, border-box, border-box, border-box, padding-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, 10%, left top, 10%, left top, 5px bottom, left top, left top, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

pl {
  background-clip: content-box, padding-box, border-box, border-box, content-box, content-box, border-box, border-box, padding-box, padding-box, border-box;
  background-color: currentColor;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, border-box, border-box, content-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: left top, left top, left top, 5px bottom, 5px bottom, 5px bottom, 5px bottom, 5px bottom, 5px bottom, 5px bottom, 5px bottom;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, 10%, cover;
}

ql {
  background-clip: content-box, border-box, content-box, border-box, border-box, content-box, content-box;
  background-color: currentColor;
  background-image: none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, border-box, border-box, border-box, border-box, border-box, content-box;
  background-position: left top, center, left top, left top, left top, 10%, 10%;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, cover;
}

rl {
  background-clip: padding-box, border-box;
  background-color: currentColor;
  background-image: none, none;
  background-origin: padding-box, content-box;
  background-position: left top, 5px bottom;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

sl {
  background-clip: padding-box, border-box, content-box, border-box, padding-box, border-box, padding-box, content-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */, none, none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, content-box, border-box, padding-box, padding-box, padding-box, content-box;
  background-position: left top, center, left top, left top, 10%, 5px bottom, left top, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, cover;
}

tl {
  background-clip: content-box, padding-box, border-box, padding-box, content-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: border-box, padding-box, border-box, padding-box, content-box;
  background-position: left top, 10%, 10%, 5px bottom, center;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, cover;
}

ul {
  background-clip: content-box, border-box, border-box, content-box, padding-box, border-box, padding-box, border-box, border-box, content-box, border-box, border-box, border-box, content-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, padding-box, border-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, content-box, padding-box, border-box, padding-box, content-box;
  background-position: 5px bottom, left top, 10%, center, left top, left top, 10%, center, left top, center, 10%, 10%, 10%, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

vl {
  background-clip: border-box, border-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box;
  background-position: left top, center, left top;
  background-repeat: repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto;
}

wl {
  background-clip: padding-box, padding-box, content-box, border-box, border-box, border-box, padding-box, content-box, content-box, content-box, padding-box, content-box, padding-box, border-box, padding-box, content-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, none /* FIXME */, none, none, none /* FIXME */, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, content-box, padding-box, border-box, content-box, padding-box, padding-box, content-box, content-box, padding-box, content-box, content-box, padding-box, padding-box, content-box, padding-box, padding-box;
  background-position: 10%, left top, 5px bottom, 10%, left top, center, center, left top, 10%, left top, 10%, left top, 5px bottom, left top, center, left top, 5px bottom, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, cover, auto, auto, auto, 10%, auto, auto, auto, cover, auto, 10%, auto, auto, auto, auto, auto;
}

xl {
  background-clip: border-box, padding-box, padding-box, border-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, border-box, border-box, padding-box;
  background-position: 5px bottom, center, left top, 10%, center;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, cover, auto;
}

yl {
  background-clip: content-box, content-box, padding-box, border-box, border-box, border-box, border-box, border-box;
  background-color: @home;
  background-image: none, none, none /* FIXME */, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, content-box, padding-box, padding-box, border-box, content-box, padding-box, border-box;
  background-position: 10%, left top, left top, 5px bottom, left top, 5px bottom, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto;
}

zl {
  background-clip: content-box, border-box, border-box, padding-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none;
  background-origin: border-box, border-box, padding-box, padding-box, border-box, content-box, padding-box;
  background-position: 10%, left top, 10%, center, left top, left top, 10%;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

am {
  background-clip: padding-box, border-box, border-box, border-box, content-box, border-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none, none, none /* FIXME */, none, none;
  background-origin: padding-box, border-box, content-box, border-box, content-box, border-box, padding-box;
  background-position: 5px bottom, center, 5px bottom, left top, 10%, center, center;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: 10%, auto, auto, auto, auto, auto, auto;
}

bm {
  background-clip: content-box, padding-box, content-box, content-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, content-box, content-box, padding-box;
  background-position: left top, left top, center, 10%;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, cover;
}

cm {
  background-clip: border-box, padding-box, border-box, content-box, border-box, border-box, content-box, content-box, padding-box, content-box, border-box, border-box, content-box, border-box, border-box, content-box, border-box, content-box, border-box, border-box, border-box, border-box;
  background-color: currentColor;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: padding-box, border-box, padding-box, content-box, padding-box, padding-box, content-box, padding-box, border-box, content-box, padding-box, border-box, border-box, padding-box, padding-box, border-box, padding-box, content-box, border-box, border-box, padding-box, border-box;
  background-position: left top, left top, center, 10%, left top, left top, center, left top, 10%, left top, 10%, left top, center, 5px bottom, center, left top, center, 10%, 5px bottom, 5px bottom, left top, 10%;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto, cover, auto, auto, auto, cover, auto, auto, auto, 10%, auto, auto, auto;
}

dm {
  background-clip: padding-box, border-box, border-box, border-box, border-box, padding-box, content-box, padding-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none, none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, border-box, padding-box, padding-box, border-box, content-box, border-box, content-box;
  background-position: center, left top, left top, left top, left top, left top, 5px bottom, left top, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, 10%, auto, auto, auto, auto, auto, auto, auto;
}

em {
  background-clip: content-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: border-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

fm {
  background-clip: border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

gm {
  background-clip: border-box, padding-box, content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, none;
  background-origin: border-box, padding-box, content-box;
  background-position: left top, 10%, left top;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, cover, auto;
}

hm {
  background-clip: content-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, content-box, border-box, content-box;
  background-position: center, left top, left top, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

im {
  background-clip: border-box, content-box, border-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, none;
  background-origin: border-box, content-box, padding-box;
  background-position: left top, left top, center;
  background-repeat: repeat, repeat, round no-repeat;
  background-size: auto, auto, cover;
}

jm {
  background-clip: border-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none, none;
  background-origin: border-box, padding-box, padding-box;
  background-position: left top, center, center;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: auto, auto, auto;
}

km {
  background-clip: content-box;
  background-color: initial;
  background-image: none;
  background-origin: content-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: auto;
}

lm {
  background-clip: content-box, content-box, content-box, padding-box, border-box, border-box, padding-box, content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */, none, none;
  background-origin: content-box, content-box, content-box, padding-box, border-box, border-box, padding-box, content-box;
  background-position: left top, 10%, left top, left top, 10%, 10%, center, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, 10%, auto, auto, auto;
}

mm {
  background-clip: border-box, content-box, content-box, border-box, border-box, padding-box, border-box, border-box, content-box, border-box, border-box, padding-box, padding-box, content-box, padding-box, padding-box, content-box, padding-box, padding-box, border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */, none;
  background-origin: padding-box, content-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box;
  background-position: left top, 5px bottom, 5px bottom, left top, left top, left top, left top, center, left top, 10%, 10%, 5px bottom, 5px bottom, left top, 10%, left top, left top, center, 5px bottom, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, cover, cover, auto, auto, 10%, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto, auto;
}

nm {
  background-clip: border-box, padding-box, content-box, border-box, content-box, content-box, border-box, border-box, content-box, border-box, padding-box, content-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none /* FIXME */, none, none /* FIXME */, none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, content-box, border-box, content-box, content-box, padding-box, padding-box, content-box, border-box, border-box, content-box;
  background-position: center, left top, 10%, left top, left top, 10%, center, 10%, left top, left top, 5px bottom, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, 10%, auto, auto, auto, cover, auto, auto, auto, auto, auto, auto;
}

om {
  background-clip: border-box;
  background-color: @home;
  background-image: none;
  background-origin: border-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

pm {
  background-clip: content-box, border-box, border-box, content-box, border-box, border-box, padding-box, border-box, border-box, content-box, border-box, border-box, content-box, border-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, border-box, border-box, border-box, border-box, padding-box, padding-box, content-box, border-box, padding-box, content-box, padding-box;
  background-position: left top, left top, 10%, center, center, 10%, center, center, center, left top, left top, 10%, left top, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, cover, auto, 10%, auto, auto, auto, cover, auto, auto, auto, cover, auto, auto;
}

qm {
  background-clip: border-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: border-box, padding-box, padding-box, content-box, padding-box, border-box, border-box, padding-box;
  background-position: 10%, 10%, left top, center, left top, left top, center, center;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, 10%, cover, auto, cover, cover, auto, auto;
}

rm {
  background-clip: content-box, padding-box, content-box, content-box, content-box, border-box, padding-box, content-box, border-box, border-box, padding-box, border-box, border-box, content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none, none, none, none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none;
  background-origin: content-box, border-box, content-box, content-box, border-box, border-box, content-box, content-box, padding-box, border-box, padding-box, padding-box, padding-box, content-box, padding-box;
  background-position: 5px bottom, left top, left top, left top, left top, 10%, center, left top, left top, center, 10%, left top, left top, left top, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, cover, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto, 10%, auto, auto;
}

sm {
  background-clip: border-box, content-box, border-box, padding-box, border-box, border-box, padding-box, padding-box, border-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, content-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box;
  background-position: 10%, left top, center, center, left top, 10%, left top, center, 10%, 10%;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, 10%, auto, auto, cover, auto, auto;
}

tm {
  background-clip: padding-box, border-box, border-box, padding-box, padding-box, border-box, content-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, padding-box, padding-box, padding-box, border-box, content-box;
  background-position: center, left top, 10%, center, 10%, 5px bottom, left top;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, cover, auto, auto, 10%, auto;
}

um {
  background-clip: padding-box, border-box, content-box, content-box, content-box, padding-box, border-box, content-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, border-box, content-box, border-box, content-box, padding-box, content-box;
  background-position: 5px bottom, 5px bottom, left top, center, 10%, left top, left top, left top;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto;
}

vm {
  background-clip: content-box, content-box, padding-box, content-box, padding-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: content-box, content-box, border-box, content-box, padding-box, padding-box;
  background-position: 10%, 10%, left top, center, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

wm {
  background-clip: padding-box, border-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box;
  background-position: 10%, left top, left top;
  background-repeat: round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto;
}

xm {
  background-clip: border-box, padding-box, border-box, border-box, border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: padding-box, padding-box, border-box, border-box, border-box, padding-box;
  background-position: 5px bottom, left top, center, 5px bottom, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

ym {
  background-clip: border-box, border-box, border-box, content-box, border-box, padding-box, content-box, border-box, border-box, border-box, padding-box, border-box, content-box, border-box, content-box, content-box, content-box, padding-box, padding-box, border-box, content-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box, border-box, content-box, padding-box, border-box, border-box, padding-box, content-box, content-box, border-box, padding-box, border-box, padding-box, border-box, border-box, border-box, border-box, padding-box, padding-box, content-box, padding-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none, none, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box, padding-box, content-box, padding-box, padding-box, border-box, border-box, border-box, border-box, padding-box, border-box, content-box, border-box, content-box, padding-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, content-box, padding-box, border-box, padding-box, border-box, border-box, border-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: 10%, left top, left top, 5px bottom, left top, left top, 10%, center, center, 10%, center, left top, 5px bottom, 10%, center, 5px bottom, 10%, left top, left top, center, 10%, left top, 5px bottom, 5px bottom, left top, left top, left top, center, center, center, center, center, left top, left top, left top, left top, 5px bottom, 5px bottom, 10%, 10%, center, 10%, center, center, 5px bottom, left top, left top;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: 10%, auto, auto, auto, auto, cover, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto, auto, auto, auto, cover, auto, 10%, auto, auto, auto, 10%, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, 10%, auto, auto, auto;
}

zm {
  background-clip: border-box, border-box, padding-box, border-box, padding-box, content-box, border-box, content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, padding-box, border-box, padding-box, content-box, border-box, content-box;
  background-position: 10%, 10%, left top, left top, center, left top, center, 10%;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, cover, 10%, 10%, auto, auto, auto, auto;
}

an {
  background-clip: content-box, border-box, content-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, content-box, padding-box, padding-box, padding-box;
  background-position: left top, left top, left top, left top, 5px bottom, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

bn {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: center;
  background-repeat: repeat;
  background-size: auto;
}

cn {
  background-clip: padding-box, padding-box, content-box, border-box, content-box, content-box, content-box, content-box, content-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none;
  background-origin: padding-box, padding-box, content-box, border-box, content-box, content-box, border-box, content-box, border-box;
  background-position: center, 10%, center, left top, left top, 10%, center, left top, center;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, 10%, auto, 10%, auto, auto, auto, auto, 10%;
}

dn {
  background-clip: border-box, content-box, content-box, padding-box, padding-box, border-box, content-box, padding-box, border-box, padding-box, content-box, padding-box, content-box, padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none;
  background-origin: padding-box, border-box, padding-box, content-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box, border-box, padding-box;
  background-position: center, left top, 10%, center, 10%, 10%, 10%, left top, left top, left top, left top, left top, 5px bottom, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

en {
  background-clip: border-box;
  background-color: rgb(255,0,0);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

fn {
  background-clip: border-box, padding-box, padding-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, padding-box, content-box;
  background-position: left top, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, cover;
}

gn {
  background-clip: content-box, padding-box, padding-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, padding-box, padding-box, border-box;
  background-position: center, center, center, 10%;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, auto;
}

hn {
  background-clip: content-box, content-box, padding-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, content-box, padding-box;
  background-position: center, left top, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: auto, auto, auto;
}

in {
  background-clip: border-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

jn {
  background-clip: content-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none, none /* FIXME */;
  background-origin: content-box, padding-box;
  background-position: center, 5px bottom;
  background-repeat: repeat, round no-repeat;
  background-size: auto, auto;
}

kn {
  background-clip: border-box, border-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: padding-box, content-box, padding-box;
  background-position: center, center, 10%;
  background-repeat: repeat, repeat, repeat;
  background-size: cover, 10%, cover;
}

ln {
  background-clip: content-box, content-box, content-box, border-box, padding-box, border-box, border-box, border-box, content-box;
  background-color: @home;
  background-image: none /* FIXME */, none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none;
  background-origin: content-box, content-box, content-box, padding-box, padding-box, border-box, padding-box, padding-box, content-box;
  background-position: center, left top, center, center, 10%, 10%, 10%, center, 10%;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, cover;
}

mn {
  background-clip: content-box, content-box, border-box, content-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none /* FIXME */, none, none;
  background-origin: padding-box, padding-box, border-box, content-box, border-box;
  background-position: center, center, left top, left top, center;
  background-repeat: repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: cover, auto, auto, auto, auto;
}

nn {
  background-clip: content-box, padding-box, border-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, padding-box, padding-box, padding-box;
  background-position: left top, left top, 5px bottom, 10%;
  background-repeat: round no-repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, 10%, auto;
}

on {
  background-clip: content-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, none /* FIXME */;
  background-origin: content-box, content-box;
  background-position: left top, 10%;
  background-repeat: round no-repeat, round no-repeat;
  background-size: auto, auto;
}

pn {
  background-clip: content-box, padding-box, padding-box, content-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, content-box;
  background-position: 5px bottom, left top, center, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, auto;
}

qn {
  background-clip: border-box, border-box, border-box, border-box, content-box, padding-box, padding-box;
  background-color: @home;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none, none;
  background-origin: padding-box, padding-box, padding-box, border-box, content-box, padding-box, border-box;
  background-position: 10%, left top, center, 10%, left top, 10%, 10%;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, 10%, auto;
}

rn {
  background-clip: padding-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */;
  background-origin: padding-box, padding-box;
  background-position: left top, 10%;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

sn {
  background-clip: padding-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box;
  background-position: center, 10%;
  background-repeat: repeat, repeat;
  background-size: 10%, cover;
}

tn {
  background-clip: border-box, border-box, border-box, border-box, padding-box, content-box, border-box, padding-box, content-box, border-box, border-box, padding-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none, none, none, none, none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, content-box, border-box, content-box, padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: 10%, 5px bottom, center, 5px bottom, 5px bottom, 5px bottom, left top, center, center, left top, left top, 5px bottom, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, 10%, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

un {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: border-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

vn {
  background-clip: border-box, border-box, content-box, border-box, padding-box, content-box, border-box, border-box, border-box, content-box;
  background-color: @home;
  background-image: none, none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, content-box, border-box, content-box, padding-box, border-box, padding-box, border-box, content-box;
  background-position: 5px bottom, 10%, center, 5px bottom, 10%, 5px bottom, center, left top, left top, left top;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, cover, auto, auto, auto, auto, auto;
}

wn {
  background-clip: padding-box, border-box, border-box, border-box, padding-box, padding-box, border-box, border-box, border-box, content-box, border-box, padding-box, border-box, content-box, padding-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none /* FIXME */, none, none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, border-box, content-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box, border-box, content-box, padding-box, border-box;
  background-position: center, center, 5px bottom, 5px bottom, left top, 10%, left top, center, center, 10%, 10%, 10%, left top, 5px bottom, center, left top;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: 10%, auto, 10%, auto, auto, auto, auto, cover, auto, cover, auto, auto, auto, auto, auto, auto;
}

xn {
  background-clip: padding-box, padding-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, border-box;
  background-position: left top, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

yn {
  background-clip: border-box, content-box, border-box, border-box, padding-box, content-box, content-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none;
  background-origin: padding-box, content-box, padding-box, padding-box, padding-box, content-box, border-box, border-box, border-box;
  background-position: left top, 5px bottom, center, 10%, center, left top, left top, left top, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, cover, auto, auto, auto, auto, auto;
}

zn {
  background-clip: border-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: none, none, none, none;
  background-origin: padding-box, border-box, padding-box, padding-box;
  background-position: left top, 10%, center, left top;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, 10%, auto, auto;
}

ao {
  background-clip: border-box, padding-box, border-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none;
  background-origin: content-box, padding-box, border-box;
  background-position: left top, center, left top;
  background-repeat: round no-repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

bo {
  background-clip: border-box, padding-box, border-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, border-box, padding-box;
  background-position: 10%, left top, left top, center;
  background-repeat: round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

co {
  background-clip: border-box, content-box, border-box, border-box, content-box, content-box, padding-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none /* FIXME */;
  background-origin: border-box, content-box, border-box, padding-box, content-box, content-box, padding-box, border-box;
  background-position: left top, 10%, center, left top, center, 5px bottom, 10%, 10%;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, 10%;
}

do {
  background-clip: padding-box, content-box;
  background-color: initial;
  background-image: none, none;
  background-origin: padding-box, content-box;
  background-position: left top, left top;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

eo {
  background-clip: border-box, content-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, content-box, border-box;
  background-position: left top, left top, left top;
  background-repeat: round no-repeat, repeat, repeat;
  background-size: 10%, auto, auto;
}

fo {
  background-clip: content-box, border-box, padding-box, border-box, border-box, border-box, padding-box, padding-box, border-box, border-box, border-box, border-box, content-box;
  background-color: initial;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none;
  background-origin: content-box, border-box, padding-box, padding-box, border-box, padding-box, border-box, padding-box, padding-box, padding-box, padding-box, border-box, content-box;
  background-position: 10%, left top, 5px bottom, left top, left top, left top, left top, left top, left top, 5px bottom, 5px bottom, left top, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

go {
  background-clip: border-box, border-box;
  background-color: @home;
  background-image: none, none;
  background-origin: border-box, border-box;
  background-position: left top, left top;
  background-repeat: round no-repeat, repeat;
  background-size: cover, auto;
}

ho {
  background-clip: border-box, padding-box, content-box, content-box, padding-box, border-box, border-box, content-box, content-box, border-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none, none /* FIXME */, none, none, none;
  background-origin: border-box, padding-box, content-box, border-box, padding-box, border-box, content-box, content-box, padding-box, border-box;
  background-position: 10%, left top, 10%, 5px bottom, 5px bottom, left top, left top, left top, 10%, center;
  background-repeat: repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, 10%;
}

io {
  background-clip: padding-box, border-box, content-box, padding-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: padding-box, content-box, border-box, padding-box;
  background-position: center, left top, 5px bottom, 5px bottom;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto;
}

jo {
  background-clip: padding-box, border-box, padding-box, border-box, padding-box, content-box, padding-box, content-box, content-box, padding-box, content-box, border-box, border-box, padding-box, padding-box, border-box, border-box;
  background-color: @home;
  background-image: none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: content-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, border-box, padding-box;
  background-position: center, 10%, center, center, left top, center, left top, center, center, left top, left top, left top, left top, center, left top, center, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, auto, cover, auto, auto, auto, cover, auto, auto, auto, auto, auto, auto, cover;
}

ko {
  background-clip: content-box, padding-box, border-box, content-box, padding-box, border-box, content-box, padding-box, padding-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, border-box, border-box, padding-box, padding-box, content-box, border-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, 10%, center, left top, 5px bottom, left top, left top, left top, left top, center, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

lo {
  background-clip: content-box, border-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none;
  background-origin: content-box, border-box;
  background-position: left top, 5px bottom;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

mo {
  background-clip: border-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box;
  background-position: left top, 10%;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

no {
  background-clip: border-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: auto;
}

oo {
  background-clip: border-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */;
  background-origin: border-box, border-box;
  background-position: left top, center;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

po {
  background-clip: border-box, border-box, padding-box, padding-box, border-box, padding-box, content-box, content-box, border-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none /* FIXME */;
  background-origin: border-box, content-box, content-box, padding-box, padding-box, padding-box, content-box, content-box, padding-box;
  background-position: 10%, 5px bottom, 5px bottom, left top, left top, 5px bottom, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, cover, auto, auto, auto, auto, auto, auto;
}

qo {
  background-clip: border-box, content-box, content-box, border-box, content-box, border-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, content-box, padding-box, border-box, content-box, padding-box;
  background-position: 10%, left top, 10%, left top, 10%, center;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

ro {
  background-clip: border-box, padding-box, border-box, border-box, content-box, content-box;
  background-color: currentColor;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: border-box, padding-box, border-box, padding-box, border-box, content-box;
  background-position: center, 5px bottom, left top, left top, left top, center;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: cover, auto, auto, auto, auto, auto;
}

so {
  background-clip: padding-box, border-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */;
  background-origin: padding-box, padding-box;
  background-position: left top, left top;
  background-repeat: round no-repeat, round no-repeat;
  background-size: auto, auto;
}

to {
  background-clip: border-box, border-box, border-box;
  background-color: @home;
  background-image: none, none, none;
  background-origin: padding-box, padding-box, padding-box;
  background-position: left top, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

uo {
  background-clip: content-box, border-box, padding-box, border-box, padding-box, border-box, padding-box, padding-box, border-box, border-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none, none /* FIXME */, none, none;
  background-origin: content-box, padding-box, padding-box, border-box, border-box, border-box, content-box, padding-box, padding-box, padding-box;
  background-position: left top, 10%, 5px bottom, left top, 10%, center, left top, left top, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto;
}

vo {
  background-clip: border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box;
  background-position: left top, 5px bottom;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

wo {
  background-clip: content-box, border-box, border-box, border-box, padding-box, content-box, border-box, border-box, border-box, border-box, border-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none /* FIXME */;
  background-origin: padding-box, border-box, padding-box, padding-box, padding-box, content-box, border-box, padding-box, content-box, border-box, padding-box, content-box, border-box;
  background-position: 5px bottom, center, left top, left top, left top, center, 5px bottom, left top, left top, left top, left top, 10%, left top;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

xo {
  background-clip: border-box, border-box, content-box, border-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none;
  background-origin: border-box, padding-box, content-box, padding-box, padding-box, content-box, border-box;
  background-position: 10%, 5px bottom, 5px bottom, left top, center, 10%, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, 10%, auto;
}

yo {
  background-clip: padding-box, border-box, padding-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, border-box;
  background-position: left top, 5px bottom, left top, center;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto;
}

zo {
  background-clip: border-box, content-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, content-box, content-box;
  background-position: 10%, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, cover;
}

ap {
  background-clip: content-box, border-box, border-box, border-box, padding-box;
  background-color: currentColor;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box, content-box, border-box, padding-box;
  background-position: left top, 10%, left top, 10%, 10%;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, cover, auto;
}

bp {
  background-clip: padding-box, border-box, border-box, padding-box, padding-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, none, none /* FIXME */, none;
  background-origin: padding-box, border-box, border-box, padding-box, padding-box;
  background-position: left top, 5px bottom, center, 5px bottom, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, cover, auto, auto;
}

cp {
  background-clip: content-box, content-box, border-box, content-box, border-box, content-box, content-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, content-box, padding-box, content-box, border-box, border-box, content-box, border-box;
  background-position: 5px bottom, 10%, left top, 10%, center, 5px bottom, left top, center;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, 10%, auto, cover, 10%, auto, auto, auto;
}

dp {
  background-clip: border-box, border-box, padding-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, center, left top, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

ep {
  background-clip: content-box, border-box, padding-box;
  background-color: currentColor;
  background-image: none, none, none;
  background-origin: border-box, border-box, padding-box;
  background-position: 5px bottom, center, center;
  background-repeat: round no-repeat, repeat, round no-repeat;
  background-size: 10%, auto, auto;
}

fp {
  background-clip: border-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none /* FIXME */, none /* FIXME */;
  background-origin: border-box, border-box, padding-box;
  background-position: left top, center, left top;
  background-repeat: repeat, repeat, round no-repeat;
  background-size: auto, cover, auto;
}

gp {
  background-clip: content-box, content-box, border-box, padding-box, border-box, content-box, padding-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none, none, none, none, none /* FIXME */, none, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, padding-box, border-box, content-box, padding-box, padding-box;
  background-position: left top, 10%, 5px bottom, 10%, 10%, 5px bottom, center, left top;
  background-repeat: repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, auto, auto, auto, auto;
}

hp {
  background-clip: content-box;
  background-color: currentColor;
  background-image: none;
  background-origin: content-box;
  background-position: 5px bottom;
  background-repeat: round no-repeat;
  background-size: auto;
}

ip {
  background-clip: border-box, border-box, border-box, padding-box, border-box, content-box, content-box, padding-box, content-box, padding-box, content-box, border-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: padding-box, border-box, border-box, padding-box, content-box, content-box, content-box, padding-box, content-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 10%, left top, 5px bottom, left top, left top, left top, left top, 5px bottom, 5px bottom, center, left top, 5px bottom, 5px bottom, center, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, auto, auto, auto, cover, 10%, auto, auto, auto, cover, auto, cover, auto;
}

jp {
  background-clip: padding-box, padding-box, border-box, border-box, border-box, border-box, content-box, border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box, border-box, padding-box, padding-box, content-box, content-box, border-box, padding-box;
  background-position: 10%, left top, 5px bottom, 10%, left top, left top, left top, left top, 10%;
  background-repeat: repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

kp {
  background-clip: border-box, border-box, padding-box, border-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, 10%, left top, left top, center, center;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

lp {
  background-clip: padding-box, border-box, border-box, padding-box, padding-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none;
  background-origin: padding-box, padding-box, padding-box, padding-box, border-box, padding-box;
  background-position: left top, 5px bottom, center, 10%, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, cover, auto, cover, auto, auto;
}

mp {
  background-clip: content-box, content-box, content-box, padding-box, padding-box, padding-box, padding-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none, none, none /* FIXME */, none, none /* FIXME */, none, none;
  background-origin: border-box, content-box, content-box, border-box, padding-box, border-box, padding-box, border-box, padding-box, padding-box;
  background-position: center, 5px bottom, 10%, left top, left top, left top, center, 5px bottom, center, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

np {
  background-clip: border-box, padding-box, padding-box, padding-box, border-box, content-box, border-box, border-box, border-box, border-box, border-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, none, none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none;
  background-origin: padding-box, padding-box, content-box, padding-box, padding-box, content-box, border-box, border-box, border-box, padding-box, padding-box, content-box;
  background-position: left top, 10%, 5px bottom, 5px bottom, left top, 10%, center, 5px bottom, 5px bottom, left top, 10%, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

op {
  background-clip: content-box, border-box, padding-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, border-box, padding-box, border-box;
  background-position: 5px bottom, left top, left top, center;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

pp {
  background-clip: border-box, padding-box, border-box, content-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none /* FIXME */, none, none;
  background-origin: border-box, border-box, padding-box, content-box, padding-box;
  background-position: left top, 5px bottom, left top, center, 5px bottom;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto;
}

qp {
  background-clip: border-box, padding-box, border-box, padding-box, border-box, content-box, border-box, content-box, border-box, border-box, content-box, border-box, border-box, content-box, padding-box, border-box, border-box;
  background-color: @home;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, padding-box, border-box, padding-box, content-box, padding-box, content-box, padding-box, padding-box, content-box, padding-box, padding-box, content-box, padding-box, border-box, padding-box;
  background-position: left top, 5px bottom, left top, center, center, 5px bottom, center, 5px bottom, 5px bottom, 10%, left top, 5px bottom, 5px bottom, left top, left top, 5px bottom, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, cover, auto, auto, auto, 10%, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto, auto;
}

rp {
  background-clip: border-box, border-box, border-box, border-box, border-box, padding-box, border-box, padding-box, content-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none;
  background-origin: border-box, border-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, content-box, border-box;
  background-position: left top, center, left top, left top, 10%, left top, 5px bottom, center, 5px bottom, 5px bottom;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto;
}

sp {
  background-clip: border-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box;
  background-position: left top, 5px bottom;
  background-repeat: round no-repeat, repeat;
  background-size: auto, 10%;
}

tp {
  background-clip: border-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

up {
  background-clip: border-box, border-box, content-box, padding-box, content-box;
  background-color: @home;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: padding-box, border-box, border-box, padding-box, content-box;
  background-position: 10%, center, 10%, 5px bottom, left top;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, 10%, auto;
}

vp {
  background-clip: border-box, content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none;
  background-origin: padding-box, content-box;
  background-position: left top, 10%;
  background-repeat: round no-repeat, repeat;
  background-size: 10%, auto;
}

wp {
  background-clip: content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

xp {
  background-clip: padding-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: padding-box;
  background-position: center;
  background-repeat: repeat;
  background-size: 10%;
}

yp {
  background-clip: border-box, border-box;
  background-color: @home;
  background-image: none /* FIXME */, none;
  background-origin: border-box, padding-box;
  background-position: 5px bottom, left top;
  background-repeat: repeat, round no-repeat;
  background-size: cover, auto;
}

zp {
  background-clip: content-box, border-box, content-box, padding-box, content-box, content-box, padding-box, padding-box, padding-box, padding-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none;
  background-origin: content-box, content-box, content-box, padding-box, content-box, content-box, padding-box, padding-box, content-box, padding-box;
  background-position: 5px bottom, left top, left top, left top, center, 10%, left top, 10%, 10%, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

aq {
  background-clip: padding-box, border-box;
  background-color: initial;
  background-image: none, none;
  background-origin: padding-box, border-box;
  background-position: 5px bottom, center;
  background-repeat: repeat, repeat;
  background-size: cover, auto;
}

bq {
  background-clip: border-box, content-box, border-box, content-box, padding-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, border-box, border-box, content-box, padding-box, padding-box, border-box;
  background-position: left top, 10%, center, center, center, left top, 10%, 10%;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: cover, 10%, auto, auto, cover, auto, auto, 10%;
}

cq {
  background-clip: border-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: content-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: 10%;
}

dq {
  background-clip: padding-box;
  background-color: initial;
  background-image: none /* FIXME */;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

eq {
  background-clip: content-box;
  background-color: currentColor;
  background-image: none;
  background-origin: content-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: auto;
}

fq {
  background-clip: content-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: auto;
}

gq {
  background-clip: padding-box, content-box, border-box, padding-box, border-box, content-box, border-box, border-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none, none /* FIXME */, none, none, none;
  background-origin: padding-box, content-box, border-box, padding-box, padding-box, content-box, padding-box, border-box, padding-box, padding-box, border-box;
  background-position: 5px bottom, 5px bottom, left top, left top, 10%, 5px bottom, center, center, 5px bottom, center, left top;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

hq {
  background-clip: padding-box, border-box, content-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box, border-box, border-box, padding-box, content-box, border-box;
  background-color: @home;
  background-image: none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, border-box, padding-box, border-box, padding-box, border-box, content-box, padding-box, content-box, content-box, border-box, padding-box, content-box, padding-box;
  background-position: 5px bottom, left top, left top, 5px bottom, left top, left top, left top, center, 5px bottom, left top, 5px bottom, left top, center, 5px bottom, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto, auto, auto, auto;
}

iq {
  background-clip: border-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, none;
  background-origin: padding-box, border-box, padding-box;
  background-position: left top, 10%, center;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, 10%;
}

jq {
  background-clip: border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none /* FIXME */;
  background-origin: border-box, padding-box;
  background-position: 10%, center;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

kq {
  background-clip: border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box;
  background-position: left top, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

lq {
  background-clip: border-box, border-box, border-box, border-box, border-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, padding-box, padding-box, content-box;
  background-position: 5px bottom, left top, left top, 10%, 10%;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: cover, auto, auto, auto, auto;
}

mq {
  background-clip: border-box, content-box, border-box, content-box, content-box, border-box, padding-box, padding-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none /* FIXME */;
  background-origin: border-box, content-box, border-box, content-box, content-box, padding-box, content-box, content-box, padding-box, border-box;
  background-position: 5px bottom, left top, center, 5px bottom, center, left top, left top, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, cover, auto, auto, auto, auto, auto;
}

nq {
  background-clip: border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

oq {
  background-clip: content-box, border-box, border-box, border-box, border-box, content-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, border-box, border-box, content-box, border-box;
  background-position: left top, left top, 10%, center, center, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, cover, cover, auto, auto;
}

pq {
  background-clip: border-box, padding-box, padding-box, border-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: border-box, padding-box, content-box, padding-box;
  background-position: 10%, 5px bottom, 5px bottom, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, 10%, auto;
}

qq {
  background-clip: border-box, content-box, padding-box, border-box, border-box, content-box, border-box, border-box, content-box, content-box, border-box, content-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none /* FIXME */, none, none, none, none /* FIXME */, none, none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, content-box, padding-box, border-box, border-box, padding-box, border-box, border-box, content-box, content-box, padding-box, content-box, padding-box;
  background-position: 5px bottom, center, center, left top, left top, left top, center, 10%, center, 5px bottom, 5px bottom, center, 10%;
  background-repeat: round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

rq {
  background-clip: border-box, border-box, padding-box, border-box;
  background-color: currentColor;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, content-box, padding-box;
  background-position: left top, left top, 5px bottom, center;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat;
  background-size: cover, auto, auto, auto;
}

sq {
  background-clip: content-box, border-box, content-box, padding-box, border-box, border-box, border-box, border-box, padding-box, border-box, border-box, padding-box, content-box, content-box, border-box, content-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none, none;
  background-origin: content-box, padding-box, content-box, border-box, border-box, padding-box, border-box, padding-box, padding-box, padding-box, border-box, padding-box, content-box, content-box, padding-box, content-box, padding-box;
  background-position: left top, 10%, 5px bottom, center, 10%, 10%, left top, center, 10%, left top, left top, left top, left top, center, center, left top, center;
  background-repeat: repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto, 10%, auto, auto, auto;
}

tq {
  background-clip: border-box, border-box, padding-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, border-box, border-box;
  background-position: 5px bottom, center, 5px bottom, 10%;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto;
}

uq {
  background-clip: padding-box, border-box, padding-box, content-box, content-box, border-box, padding-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, padding-box, content-box, content-box, border-box, padding-box, padding-box, border-box;
  background-position: center, center, left top, left top, left top, 5px bottom, 10%, 10%, left top;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, cover, auto, auto, auto, auto, auto, auto, auto;
}

vq {
  background-clip: padding-box, content-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none;
  background-origin: border-box, content-box, padding-box;
  background-position: center, 10%, 10%;
  background-repeat: repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto;
}

wq {
  background-clip: border-box, border-box;
  background-color: initial;
  background-image: none, none;
  background-origin: padding-box, border-box;
  background-position: 5px bottom, center;
  background-repeat: round no-repeat, round no-repeat;
  background-size: auto, auto;
}

xq {
  background-clip: content-box, border-box, border-box, padding-box, padding-box, border-box, border-box, border-box, padding-box, content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none;
  background-origin: content-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box;
  background-position: left top, left top, left top, 5px bottom, center, center, left top, 10%, left top, center;
  background-repeat: round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, 10%, auto, auto, auto;
}

yq {
  background-clip: border-box, border-box, content-box, content-box;
  background-color: currentColor;
  background-image: none, none, none, none;
  background-origin: padding-box, padding-box, content-box, padding-box;
  background-position: 5px bottom, left top, left top, 5px bottom;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, 10%, auto, auto;
}

zq {
  background-clip: content-box, padding-box, border-box, content-box, border-box, padding-box, padding-box, padding-box, padding-box, border-box, border-box, padding-box, border-box, content-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none;
  background-origin: content-box, padding-box, border-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box, border-box, border-box, content-box, padding-box, content-box, content-box;
  background-position: 5px bottom, left top, 10%, 10%, 5px bottom, left top, 5px bottom, left top, 10%, 10%, 5px bottom, left top, left top, center, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

ar {
  background-clip: border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: border-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

br {
  background-clip: border-box, content-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none;
  background-origin: padding-box, padding-box, border-box;
  background-position: 10%, 10%, left top;
  background-repeat: round no-repeat, round no-repeat, repeat;
  background-size: 10%, auto, cover;
}

cr {
  background-clip: border-box, padding-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, content-box, padding-box, padding-box, content-box, border-box, padding-box;
  background-position: left top, 10%, 5px bottom, 10%, left top, 5px bottom, left top, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto;
}

dr {
  background-clip: border-box, content-box;
  background-color: @home;
  background-image: none /* FIXME */, none;
  background-origin: padding-box, content-box;
  background-position: 10%, 5px bottom;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

er {
  background-clip: border-box, content-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none;
  background-origin: padding-box, border-box, padding-box;
  background-position: center, 10%, left top;
  background-repeat: round no-repeat, repeat, repeat;
  background-size: cover, cover, auto;
}

fr {
  background-clip: content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box;
  background-position: 10%;
  background-repeat: round no-repeat;
  background-size: 10%;
}

gr {
  background-clip: padding-box, padding-box, border-box, padding-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, border-box, content-box;
  background-position: 10%, left top, 5px bottom, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

hr {
  background-clip: border-box, border-box, content-box, content-box, border-box, border-box, border-box, border-box, content-box, border-box, border-box, padding-box, content-box, border-box, padding-box, content-box, padding-box, content-box, padding-box, border-box, border-box, content-box, border-box, border-box, border-box, border-box, padding-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none, none, none, none /* FIXME */, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: border-box, padding-box, padding-box, content-box, padding-box, border-box, padding-box, border-box, content-box, border-box, padding-box, border-box, content-box, border-box, content-box, border-box, padding-box, padding-box, padding-box, border-box, border-box, padding-box, padding-box, border-box, border-box, border-box, padding-box, padding-box, content-box;
  background-position: center, left top, left top, left top, left top, 5px bottom, center, 5px bottom, left top, 5px bottom, 5px bottom, left top, 5px bottom, left top, 5px bottom, 5px bottom, center, left top, left top, left top, left top, left top, 10%, left top, left top, 10%, left top, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, cover, auto, auto, auto, cover, cover, cover, auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto;
}

ir {
  background-clip: padding-box, border-box;
  background-color: initial;
  background-image: none, none;
  background-origin: padding-box, padding-box;
  background-position: left top, center;
  background-repeat: repeat, repeat;
  background-size: auto, cover;
}

jr {
  background-clip: border-box, padding-box, content-box, border-box, content-box, padding-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none;
  background-origin: padding-box, padding-box, content-box, border-box, content-box, border-box, border-box;
  background-position: center, 10%, center, center, left top, 10%, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: 10%, auto, auto, auto, auto, cover, cover;
}

kr {
  background-clip: content-box, padding-box, border-box, padding-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none;
  background-origin: padding-box, padding-box, border-box, padding-box, content-box, border-box, padding-box;
  background-position: left top, 10%, left top, center, 10%, left top, 10%;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

lr {
  background-clip: border-box, border-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, content-box, border-box, border-box, padding-box;
  background-position: left top, 5px bottom, 10%, 5px bottom, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto;
}

mr {
  background-clip: border-box;
  background-color: currentColor;
  background-image: none;
  background-origin: padding-box;
  background-position: center;
  background-repeat: repeat;
  background-size: auto;
}

nr {
  background-clip: content-box, padding-box, content-box, padding-box, padding-box, border-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none, none, none, none /* FIXME */, none /* FIXME */;
  background-origin: content-box, padding-box, content-box, padding-box, padding-box, border-box;
  background-position: left top, 5px bottom, 10%, left top, left top, 10%;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

or {
  background-clip: padding-box, padding-box, content-box, border-box, border-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, content-box, padding-box, padding-box;
  background-position: 5px bottom, left top, 10%, 10%, 10%;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, 10%, auto;
}

pr {
  background-clip: content-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, padding-box, border-box;
  background-position: center, left top, 10%, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto;
}

qr {
  background-clip: content-box, border-box, content-box, border-box, content-box;
  background-color: initial;
  background-image: none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: content-box, padding-box, content-box, padding-box, content-box;
  background-position: 5px bottom, center, center, 5px bottom, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat;
  background-size: auto, 10%, auto, cover, auto;
}

rr {
  background-clip: border-box, content-box, border-box, border-box, border-box, padding-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none;
  background-origin: border-box, content-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: left top, left top, left top, 10%, left top, 5px bottom, 10%, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, 10%, auto, auto;
}

sr {
  background-clip: border-box, border-box, padding-box, border-box, border-box, content-box, content-box, border-box, border-box, content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none /* FIXME */, none /* FIXME */, none, none, none;
  background-origin: padding-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box, content-box, padding-box, padding-box;
  background-position: 5px bottom, left top, center, left top, left top, left top, center, 5px bottom, center, left top;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, cover, auto, auto, auto, auto;
}

tr {
  background-clip: border-box, padding-box, content-box, content-box, content-box, border-box, content-box, border-box, padding-box, border-box, padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, content-box, content-box, border-box, border-box, border-box, border-box, padding-box, content-box, padding-box;
  background-position: left top, 10%, left top, left top, left top, 5px bottom, left top, 10%, left top, 5px bottom, 5px bottom;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, cover, auto, 10%, 10%, auto, auto, auto;
}

ur {
  background-clip: border-box, border-box, border-box;
  background-color: @home;
  background-image: none, none /* FIXME */, none;
  background-origin: padding-box, padding-box, border-box;
  background-position: left top, 5px bottom, left top;
  background-repeat: round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, 10%;
}

vr {
  background-clip: padding-box, padding-box, padding-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none;
  background-origin: padding-box, padding-box, padding-box, padding-box;
  background-position: center, left top, left top, center;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, cover;
}

wr {
  background-clip: padding-box, border-box;
  background-color: @home;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box;
  background-position: 5px bottom, 10%;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

xr {
  background-clip: content-box, border-box, border-box, padding-box, border-box, border-box, content-box, content-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, none /* FIXME */, none, none, none, none, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, padding-box, padding-box, content-box, content-box, content-box, content-box;
  background-position: 5px bottom, left top, 10%, 5px bottom, center, 10%, center, left top, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, 10%, auto, auto, auto, auto, auto, auto, auto;
}

yr {
  background-clip: border-box, border-box, padding-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none /* FIXME */, none, none;
  background-origin: padding-box, padding-box, padding-box, padding-box;
  background-position: left top, 5px bottom, left top, 10%;
  background-repeat: round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, cover;
}

zr {
  background-clip: border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box;
  background-position: left top, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

as {
  background-clip: border-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box;
  background-position: left top, center, center;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

bs {
  background-clip: content-box, padding-box, border-box, padding-box, border-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, content-box, padding-box, padding-box;
  background-position: 10%, left top, left top, 5px bottom, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, 10%, auto;
}

cs {
  background-clip: border-box, content-box, border-box, border-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, border-box, padding-box;
  background-position: left top, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

ds {
  background-clip: content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */;
  background-origin: padding-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: auto;
}

es {
  background-clip: border-box, border-box, padding-box, border-box, padding-box, border-box, border-box, border-box, border-box, padding-box, content-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */;
  background-origin: border-box, padding-box, padding-box, border-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, content-box, padding-box;
  background-position: left top, 10%, left top, 5px bottom, left top, 10%, 10%, left top, 10%, left top, 10%, center;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, cover, auto, auto, auto, cover, auto, auto, auto, auto, auto, 10%;
}

fs {
  background-clip: content-box, content-box, border-box, padding-box, border-box, padding-box, border-box, padding-box, padding-box, border-box;
  background-color: @home;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: content-box, content-box, padding-box, padding-box, padding-box, padding-box, content-box, content-box, padding-box, border-box;
  background-position: 10%, 5px bottom, 5px bottom, center, 10%, left top, center, left top, 5px bottom, 10%;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, cover, auto, auto, auto, auto;
}

gs {
  background-clip: border-box, content-box, border-box, padding-box, border-box, padding-box, border-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: border-box, content-box, padding-box, padding-box, padding-box, padding-box, border-box;
  background-position: left top, 5px bottom, 10%, left top, center, left top, 5px bottom;
  background-repeat: round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

hs {
  background-clip: content-box, border-box, content-box, border-box, content-box, border-box, content-box, border-box, border-box;
  background-color: @home;
  background-image: none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, border-box, content-box, padding-box, content-box, border-box, border-box, border-box, padding-box;
  background-position: 10%, left top, 10%, 10%, 5px bottom, 5px bottom, left top, center, 5px bottom;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, 10%, auto, auto, auto, 10%, auto;
}

is {
  background-clip: padding-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

js {
  background-clip: padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: border-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

ks {
  background-clip: content-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: border-box;
  background-position: 10%;
  background-repeat: round no-repeat;
  background-size: auto;
}

ls {
  background-clip: border-box, padding-box, padding-box, border-box;
  background-color: currentColor;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, padding-box, padding-box;
  background-position: left top, 10%, left top, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, cover, auto;
}

ms {
  background-clip: border-box, content-box, content-box, border-box;
  background-color: @home;
  background-image: none, none /* FIXME */, none, none;
  background-origin: border-box, content-box, border-box, border-box;
  background-position: left top, center, 10%, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, 10%, auto;
}

ns {
  background-clip: padding-box, border-box, padding-box, content-box, padding-box, content-box, border-box;
  background-color: @home;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none;
  background-origin: padding-box, padding-box, padding-box, padding-box, border-box, content-box, padding-box;
  background-position: left top, 5px bottom, 10%, left top, 5px bottom, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: cover, auto, 10%, auto, cover, 10%, auto;
}

os {
  background-clip: content-box, padding-box, padding-box, border-box, padding-box, border-box, padding-box, padding-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, padding-box, padding-box, content-box, padding-box, border-box, border-box, padding-box, padding-box, content-box, padding-box;
  background-position: center, left top, 5px bottom, left top, 5px bottom, left top, 10%, 10%, center, center, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, 10%, cover, auto, cover, auto, auto, cover, auto, auto, auto;
}

ps {
  background-clip: padding-box, content-box, content-box, content-box, border-box, border-box, padding-box, border-box, border-box, border-box, content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, border-box, padding-box, content-box, padding-box, border-box, padding-box, padding-box, padding-box, padding-box, content-box;
  background-position: left top, center, left top, center, left top, left top, left top, center, left top, 5px bottom, 10%;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, cover;
}

qs {
  background-clip: content-box, border-box, content-box, content-box, content-box, border-box;
  background-color: @home;
  background-image: none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box, content-box, content-box, content-box, padding-box;
  background-position: 10%, 10%, 10%, 10%, left top, center;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, cover, cover, auto, auto;
}

rs {
  background-clip: content-box, padding-box, border-box, content-box, content-box, border-box, padding-box, border-box, border-box, padding-box, padding-box;
  background-color: @home;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, padding-box, padding-box, border-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box;
  background-position: center, center, left top, left top, 5px bottom, 5px bottom, left top, center, 5px bottom, left top, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto;
}

ss {
  background-clip: border-box, padding-box, border-box, padding-box, content-box, content-box, border-box, content-box, padding-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box, padding-box, content-box, content-box, border-box, content-box, padding-box, border-box, content-box, padding-box;
  background-position: left top, center, 10%, left top, 5px bottom, left top, left top, left top, left top, left top, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, auto, auto, cover, 10%, auto, auto, auto, auto, auto;
}

ts {
  background-clip: border-box, content-box, padding-box, border-box, content-box, border-box, padding-box, padding-box, content-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: left top, left top, center, center, 5px bottom, center, left top, center, 10%, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, cover, auto, 10%, cover, auto, 10%, auto, auto;
}

us {
  background-clip: padding-box;
  background-color: currentColor;
  background-image: none;
  background-origin: padding-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: auto;
}

vs {
  background-clip: border-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

ws {
  background-clip: border-box, border-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, border-box, padding-box;
  background-position: left top, center, left top;
  background-repeat: round no-repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

xs {
  background-clip: border-box, padding-box, padding-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, padding-box, padding-box, content-box;
  background-position: 10%, 5px bottom, left top, left top, center;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat;
  background-size: cover, auto, auto, auto, auto;
}

ys {
  background-clip: border-box, border-box, border-box, padding-box, content-box, content-box, border-box, padding-box, padding-box, content-box, border-box;
  background-color: @home;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, border-box, padding-box, padding-box, content-box, content-box, padding-box, padding-box, padding-box, content-box, border-box;
  background-position: center, 10%, left top, 5px bottom, 5px bottom, 5px bottom, 5px bottom, 5px bottom, left top, 5px bottom, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

zs {
  background-clip: padding-box, border-box, padding-box, padding-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */;
  background-origin: padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: left top, left top, left top, 10%, left top;
  background-repeat: repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, cover;
}

at {
  background-clip: border-box, border-box, border-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, padding-box, padding-box, border-box;
  background-position: center, left top, center, center, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, cover, auto, auto, auto;
}

bt {
  background-clip: border-box, padding-box, content-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: border-box, content-box, content-box, border-box, border-box;
  background-position: center, left top, 10%, 5px bottom, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, cover, auto;
}

ct {
  background-clip: border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none;
  background-origin: border-box, border-box;
  background-position: left top, center;
  background-repeat: repeat, repeat;
  background-size: auto, cover;
}

dt {
  background-clip: content-box;
  background-color: initial;
  background-image: none;
  background-origin: content-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

et {
  background-clip: border-box, border-box, padding-box, padding-box, border-box, padding-box, content-box;
  background-color: @home;
  background-image: none, none, none, none /* FIXME */, none, none, none;
  background-origin: padding-box, border-box, padding-box, padding-box, padding-box, padding-box, content-box;
  background-position: left top, left top, left top, center, 10%, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: cover, auto, auto, auto, 10%, auto, auto;
}

ft {
  background-clip: border-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, content-box;
  background-position: left top, left top;
  background-repeat: repeat, repeat;
  background-size: auto, cover;
}

gt {
  background-clip: padding-box;
  background-color: rgb(255,0,0);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

ht {
  background-clip: border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none /* FIXME */;
  background-origin: padding-box, border-box;
  background-position: left top, 10%;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

it {
  background-clip: content-box, content-box, border-box, border-box, padding-box, content-box, padding-box, border-box, content-box, border-box, border-box, border-box, content-box, border-box, padding-box, border-box, border-box, border-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box, border-box, border-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, border-box, padding-box, padding-box;
  background-position: center, 5px bottom, left top, 5px bottom, left top, 5px bottom, left top, left top, left top, left top, center, left top, 5px bottom, 5px bottom, 10%, left top, left top, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, 10%, 10%, auto, auto, cover, auto, auto, auto, auto, auto, 10%, auto, cover, cover, auto, auto, auto;
}

jt {
  background-clip: border-box, border-box, border-box, padding-box, border-box, content-box, border-box, border-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box;
  background-position: 10%, left top, left top, left top, 10%, 5px bottom, 10%, left top, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, 10%, auto, auto, auto, auto;
}

kt {
  background-clip: padding-box, border-box, content-box, content-box, padding-box, border-box, border-box, border-box, content-box, content-box, border-box, content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none /* FIXME */, none, none /* FIXME */, none, none;
  background-origin: padding-box, border-box, content-box, content-box, padding-box, padding-box, padding-box, padding-box, content-box, content-box, border-box, content-box;
  background-position: left top, 5px bottom, center, center, left top, 10%, left top, left top, 10%, center, center, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, 10%, auto, auto, auto, auto, auto, cover, auto, cover, auto, auto;
}

lt {
  background-clip: border-box, content-box, border-box, border-box, border-box, content-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none;
  background-origin: padding-box, content-box, border-box, padding-box, border-box, content-box;
  background-position: center, left top, left top, left top, left top, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, 10%, auto, auto;
}

mt {
  background-clip: border-box, border-box, padding-box, border-box, content-box, border-box, border-box, border-box, padding-box, padding-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none /* FIXME */, none, none, none, none, none /* FIXME */, none /* FIXME */, none, none;
  background-origin: padding-box, content-box, border-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box;
  background-position: center, 5px bottom, 10%, left top, 10%, left top, 10%, left top, 5px bottom, center, left top;
  background-repeat: round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: cover, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

nt {
  background-clip: content-box, border-box, border-box, border-box, content-box, content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none;
  background-origin: content-box, border-box, padding-box, border-box, padding-box, content-box, padding-box;
  background-position: left top, 5px bottom, left top, left top, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

ot {
  background-clip: padding-box, border-box, border-box, padding-box, border-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none;
  background-origin: padding-box, padding-box, padding-box, padding-box, content-box;
  background-position: center, left top, 10%, left top, center;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, auto;
}

pt {
  background-clip: border-box, content-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: border-box, content-box, border-box;
  background-position: 5px bottom, center, center;
  background-repeat: repeat, repeat, round no-repeat;
  background-size: 10%, auto, cover;
}

qt {
  background-clip: content-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, none;
  background-origin: content-box, padding-box;
  background-position: left top, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

rt {
  background-clip: border-box, content-box, border-box;
  background-color: initial;
  background-image: none, none, none;
  background-origin: content-box, content-box, border-box;
  background-position: left top, center, left top;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

st {
  background-clip: content-box, padding-box, border-box, content-box, content-box, padding-box, border-box, content-box, padding-box, border-box, padding-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none /* FIXME */, none;
  background-origin: padding-box, padding-box, border-box, content-box, content-box, padding-box, padding-box, content-box, padding-box, border-box, padding-box, border-box, padding-box;
  background-position: center, 10%, 5px bottom, left top, 10%, left top, left top, left top, left top, left top, left top, 5px bottom, left top;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto, 10%, auto;
}

tt {
  background-clip: border-box, border-box, border-box, content-box, border-box, border-box, border-box, content-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box, content-box, padding-box, padding-box, border-box, content-box, border-box;
  background-position: 5px bottom, 5px bottom, left top, center, left top, left top, center, left top, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

ut {
  background-clip: padding-box, border-box, border-box, padding-box, content-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: border-box, border-box, border-box, content-box, content-box, padding-box;
  background-position: left top, left top, 5px bottom, left top, left top, 10%;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

vt {
  background-clip: content-box, padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: content-box, padding-box;
  background-position: 5px bottom, left top;
  background-repeat: round no-repeat, repeat;
  background-size: auto, cover;
}

wt {
  background-clip: border-box, content-box, border-box, border-box, border-box, padding-box, content-box, padding-box, padding-box;
  background-color: initial;
  background-image: none, none, none, none /* FIXME */, none, none, none, none, none;
  background-origin: padding-box, padding-box, padding-box, border-box, padding-box, padding-box, content-box, padding-box, padding-box;
  background-position: 5px bottom, 5px bottom, left top, 5px bottom, left top, 10%, 10%, left top, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, 10%, cover, 10%, auto, 10%, auto, auto, auto;
}

xt {
  background-clip: border-box, padding-box, border-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: border-box, border-box, padding-box, content-box;
  background-position: 10%, 5px bottom, 10%, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto;
}

yt {
  background-clip: padding-box, border-box, border-box, border-box, content-box, content-box, content-box, padding-box, padding-box, content-box, border-box, border-box, border-box, padding-box, border-box, border-box, border-box, border-box, padding-box, border-box, border-box, border-box, content-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none, none, none, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box, padding-box, content-box, content-box, content-box, padding-box, border-box, border-box, content-box, padding-box, border-box, padding-box, border-box, border-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box;
  background-position: center, left top, 10%, left top, center, 10%, center, center, left top, left top, 10%, 5px bottom, 5px bottom, 5px bottom, center, 10%, left top, left top, left top, 10%, 5px bottom, center, center;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: 10%, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto, auto, auto, auto, cover, auto;
}

zt {
  background-clip: padding-box, padding-box, border-box, padding-box, border-box, border-box, border-box, content-box, padding-box, content-box, border-box, border-box, content-box, content-box, border-box;
  background-color: currentColor;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: content-box, padding-box, border-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, border-box, border-box, padding-box, content-box, padding-box;
  background-position: 10%, left top, center, left top, left top, left top, center, center, center, center, 10%, 10%, center, left top, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: 10%, auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto;
}

au {
  background-clip: padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

bu {
  background-clip: padding-box, border-box, border-box, content-box, border-box;
  background-color: currentColor;
  background-image: none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, content-box, border-box, padding-box;
  background-position: center, left top, left top, 10%, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto;
}

cu {
  background-clip: border-box, content-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, content-box, border-box, padding-box;
  background-position: left top, 10%, center, 5px bottom, 10%;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, 10%, auto, auto, auto;
}

du {
  background-clip: border-box, content-box, padding-box, border-box, padding-box, border-box, content-box, padding-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: left top, 10%, center, left top, left top, 5px bottom, left top, 10%, left top;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, cover;
}

eu {
  background-clip: padding-box, content-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box;
  background-position: left top, center;
  background-repeat: repeat, round no-repeat;
  background-size: auto, auto;
}

fu {
  background-clip: padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */;
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

gu {
  background-clip: content-box, border-box, border-box, border-box, padding-box, border-box, content-box, content-box, border-box, content-box, border-box, content-box, border-box, content-box;
  background-color: currentColor;
  background-image: none, none, none, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, border-box, padding-box, padding-box, content-box, border-box, padding-box, content-box, border-box, content-box, border-box, content-box;
  background-position: 5px bottom, left top, left top, 10%, center, 5px bottom, left top, center, left top, 5px bottom, left top, left top, 5px bottom, center;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: cover, auto, auto, auto, auto, cover, auto, 10%, auto, auto, auto, auto, auto, 10%;
}

hu {
  background-clip: border-box, padding-box, border-box, border-box, content-box, border-box, content-box, content-box, padding-box, border-box, border-box, padding-box, content-box;
  background-color: @home;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, border-box, padding-box, border-box, content-box, padding-box, content-box, content-box, content-box, padding-box, padding-box, padding-box, content-box;
  background-position: 10%, center, center, 10%, 5px bottom, center, left top, left top, left top, 5px bottom, 10%, 5px bottom, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, 10%, auto, auto, auto, auto, auto, auto, 10%, auto, auto, auto, 10%;
}

iu {
  background-clip: content-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box;
  background-position: 5px bottom, 5px bottom;
  background-repeat: repeat, repeat;
  background-size: auto, cover;
}

ju {
  background-clip: content-box;
  background-color: rgb(255,0,0);
  background-image: none;
  background-origin: content-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: auto;
}

ku {
  background-clip: border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */;
  background-origin: padding-box;
  background-position: 10%;
  background-repeat: round no-repeat;
  background-size: auto;
}

lu {
  background-clip: padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

mu {
  background-clip: content-box, content-box, padding-box, content-box, border-box, padding-box, padding-box, content-box, border-box, content-box, content-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, content-box, border-box, padding-box, padding-box, content-box, content-box, padding-box, border-box;
  background-position: left top, 10%, left top, 10%, center, left top, center, center, left top, left top, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: 10%, auto, 10%, auto, 10%, auto, auto, auto, auto, auto, auto;
}

nu {
  background-clip: padding-box, content-box, padding-box, content-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, content-box, padding-box, border-box;
  background-position: 10%, center, 5px bottom, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, round no-repeat;
  background-size: cover, auto, cover, auto;
}

ou {
  background-clip: padding-box, padding-box, padding-box, border-box, padding-box, content-box, border-box, padding-box, border-box, border-box;
  background-color: @home;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, padding-box, border-box, padding-box, content-box, padding-box, padding-box, padding-box, border-box;
  background-position: center, center, 10%, left top, 5px bottom, 5px bottom, center, left top, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, cover, cover, auto, auto;
}

pu {
  background-clip: padding-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */;
  background-origin: padding-box;
  background-position: center;
  background-repeat: repeat;
  background-size: auto;
}

qu {
  background-clip: border-box, border-box, padding-box, border-box, padding-box, border-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none, none, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, border-box, border-box, padding-box;
  background-position: left top, 5px bottom, center, 5px bottom, left top, center, 10%, 5px bottom;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, 10%, auto, auto, auto, auto, 10%, auto;
}

ru {
  background-clip: content-box, content-box, border-box, content-box, border-box, border-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, content-box, padding-box, content-box, padding-box, border-box, padding-box, border-box;
  background-position: 5px bottom, 5px bottom, 5px bottom, center, left top, left top, 10%, 10%;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, 10%, auto, 10%, auto, auto;
}

su {
  background-clip: border-box, border-box, content-box, border-box, border-box, padding-box, padding-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 10%, left top, 10%, 5px bottom, 5px bottom, left top, 10%, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, 10%, auto, auto, auto, cover;
}

tu {
  background-clip: border-box, border-box, border-box, padding-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box, padding-box, border-box, padding-box;
  background-position: center, 10%, 10%, center, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: cover, cover, auto, auto, auto;
}

uu {
  background-clip: border-box, padding-box, border-box, border-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, content-box, padding-box;
  background-position: center, 5px bottom, 5px bottom, left top, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat;
  background-size: auto, cover, cover, auto, auto;
}

vu {
  background-clip: border-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, none, none;
  background-origin: border-box, content-box, padding-box, content-box;
  background-position: center, left top, 5px bottom, center;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, cover;
}

wu {
  background-clip: border-box, content-box, content-box, border-box, content-box, border-box, padding-box, padding-box;
  background-color: currentColor;
  background-image: none, none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: border-box, content-box, content-box, padding-box, content-box, border-box, border-box, padding-box;
  background-position: left top, left top, center, left top, left top, 10%, 5px bottom, 5px bottom;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, 10%, auto, 10%;
}

xu {
  background-clip: border-box, padding-box, content-box, content-box;
  background-color: @home;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: border-box, padding-box, content-box, content-box;
  background-position: left top, center, 10%, left top;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, 10%, auto, auto;
}

yu {
  background-clip: border-box, content-box, content-box, content-box, padding-box, padding-box, border-box, content-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none, none /* FIXME */, none, none, none, none, none;
  background-origin: border-box, content-box, padding-box, content-box, padding-box, padding-box, padding-box, content-box;
  background-position: 10%, left top, center, 10%, 5px bottom, 5px bottom, 10%, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: cover, auto, auto, auto, auto, auto, auto, auto;
}

zu {
  background-clip: content-box;
  background-color: currentColor;
  background-image: none;
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: round no-repeat;
  background-size: auto;
}

av {
  background-clip: border-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: round no-repeat;
  background-size: cover;
}

bv {
  background-clip: content-box, padding-box, border-box, padding-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, border-box, padding-box, padding-box, border-box, padding-box;
  background-position: 10%, left top, left top, 10%, 5px bottom, 10%, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, cover, auto;
}

cv {
  background-clip: padding-box, padding-box, border-box, border-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none /* FIXME */, none, none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: left top, left top, 10%, left top, center, center;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, 10%, auto;
}

dv {
  background-clip: content-box, border-box, padding-box, border-box, padding-box, content-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none, none, none;
  background-origin: content-box, border-box, border-box, border-box, padding-box, content-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box;
  background-position: left top, 5px bottom, 5px bottom, 10%, 5px bottom, center, left top, left top, 10%, 5px bottom, center, left top, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

ev {
  background-clip: content-box;
  background-color: rgb(255,0,0);
  background-image: none;
  background-origin: content-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: cover;
}

fv {
  background-clip: border-box, content-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, padding-box;
  background-position: left top, left top, 5px bottom;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

gv {
  background-clip: border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: center;
  background-repeat: repeat;
  background-size: cover;
}

hv {
  background-clip: border-box, content-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */;
  background-origin: padding-box, content-box;
  background-position: center, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

iv {
  background-clip: border-box, content-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, padding-box;
  background-position: left top, center, left top;
  background-repeat: repeat, repeat, round no-repeat;
  background-size: auto, auto, auto;
}

jv {
  background-clip: border-box, content-box, content-box, content-box, border-box, border-box, padding-box, padding-box, border-box, content-box, border-box, border-box, border-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: padding-box, content-box, border-box, content-box, content-box, border-box, padding-box, padding-box, padding-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: left top, center, center, left top, center, 10%, 10%, 5px bottom, 5px bottom, center, 5px bottom, 10%, 5px bottom, left top, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, 10%, auto, auto, auto, cover, auto, auto, auto, 10%, auto, cover, auto;
}

kv {
  background-clip: border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

lv {
  background-clip: border-box, content-box, padding-box, content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, content-box, content-box, content-box;
  background-position: 5px bottom, center, center, center;
  background-repeat: repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto;
}

mv {
  background-clip: content-box, padding-box, content-box, border-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, border-box, border-box;
  background-position: left top, left top, center, center, 10%;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto;
}

nv {
  background-clip: padding-box, content-box, content-box, padding-box, padding-box, content-box, border-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, none /* FIXME */;
  background-origin: padding-box, content-box, border-box, padding-box, padding-box, content-box, padding-box;
  background-position: center, left top, 10%, left top, left top, left top, 5px bottom;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat;
  background-size: 10%, cover, cover, auto, auto, auto, auto;
}

ov {
  background-clip: content-box, border-box, border-box, content-box, content-box;
  background-color: initial;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, padding-box, border-box, content-box;
  background-position: left top, center, 10%, 5px bottom, 10%;
  background-repeat: repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: cover, 10%, auto, auto, auto;
}

pv {
  background-clip: border-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: padding-box, padding-box;
  background-position: center, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

qv {
  background-clip: border-box, border-box, border-box, border-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-position: left top, 5px bottom, center, left top, left top, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, 10%, auto, cover;
}

rv {
  background-clip: border-box, border-box, content-box, border-box, border-box, padding-box, border-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */, none;
  background-origin: padding-box, padding-box, border-box, border-box, padding-box, padding-box, padding-box, border-box;
  background-position: 10%, left top, left top, left top, center, left top, left top, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, 10%, auto;
}

sv {
  background-clip: border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none;
  background-origin: padding-box, border-box;
  background-position: 10%, 10%;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

tv {
  background-clip: border-box, content-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none;
  background-origin: content-box, content-box, padding-box;
  background-position: left top, left top, left top;
  background-repeat: round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto;
}

uv {
  background-clip: border-box, content-box;
  background-color: @home;
  background-image: none, none;
  background-origin: padding-box, content-box;
  background-position: left top, left top;
  background-repeat: repeat, repeat;
  background-size: auto, cover;
}

vv {
  background-clip: padding-box, border-box, padding-box, border-box, border-box, border-box, content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none, none /* FIXME */, none, none, none /* FIXME */, none;
  background-origin: border-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box;
  background-position: left top, 5px bottom, left top, center, center, 5px bottom, left top;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

wv {
  background-clip: content-box, content-box, border-box, border-box, content-box, border-box;
  background-color: currentColor;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, padding-box, padding-box, content-box, padding-box;
  background-position: left top, left top, center, 10%, left top, center;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat;
  background-size: cover, auto, 10%, auto, cover, auto;
}

xv {
  background-clip: border-box, padding-box, border-box, padding-box, border-box, content-box, border-box, content-box, border-box;
  background-color: @home;
  background-image: none, none, none /* FIXME */, none, none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, content-box, border-box, content-box, padding-box;
  background-position: 5px bottom, 5px bottom, left top, left top, left top, center, 10%, left top, 10%;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, cover, auto, auto, auto;
}

yv {
  background-clip: border-box, padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, content-box;
  background-position: 5px bottom, left top;
  background-repeat: round no-repeat, repeat;
  background-size: cover, auto;
}

zv {
  background-clip: padding-box, border-box;
  background-color: @home;
  background-image: none /* FIXME */, none;
  background-origin: border-box, border-box;
  background-position: left top, left top;
  background-repeat: repeat, round no-repeat;
  background-size: auto, auto;
}

aw {
  background-clip: content-box, content-box, padding-box;
  background-color: @home;
  background-image: none, none /* FIXME */, none;
  background-origin: content-box, content-box, padding-box;
  background-position: left top, left top, center;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: auto, auto, auto;
}

bw {
  background-clip: padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box;
  background-position: center;
  background-repeat: repeat;
  background-size: auto;
}

cw {
  background-clip: border-box, border-box, content-box, border-box, border-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, content-box, padding-box, border-box, padding-box;
  background-position: 10%, 10%, 5px bottom, left top, left top, center;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, 10%, auto, auto, auto, auto;
}

dw {
  background-clip: border-box, content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, content-box;
  background-position: 10%, center;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

ew {
  background-clip: border-box, content-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none;
  background-origin: padding-box, content-box, padding-box;
  background-position: 10%, left top, 10%;
  background-repeat: round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, 10%;
}

fw {
  background-clip: border-box, padding-box, border-box, content-box;
  background-color: rgb(170,221,255);
  background-image: none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, border-box, padding-box, content-box;
  background-position: center, left top, 10%, 10%;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

gw {
  background-clip: border-box, border-box, padding-box;
  background-color: currentColor;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, padding-box;
  background-position: left top, 10%, 10%;
  background-repeat: repeat, repeat, repeat;
  background-size: cover, 10%, auto;
}

hw {
  background-clip: content-box, content-box, content-box, content-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, none;
  background-origin: content-box, content-box, content-box, border-box;
  background-position: left top, left top, center, left top;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto;
}

iw {
  background-clip: border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none;
  background-origin: padding-box, border-box;
  background-position: left top, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

jw {
  background-clip: border-box;
  background-color: currentColor;
  background-image: none;
  background-origin: border-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: auto;
}

kw {
  background-clip: border-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: border-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box;
  background-position: 10%, 10%, 10%, center, 10%, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, 10%;
}

lw {
  background-clip: padding-box, content-box, padding-box, border-box;
  background-color: @home;
  background-image: none, none /* FIXME */, none, none;
  background-origin: padding-box, content-box, padding-box, border-box;
  background-position: center, 10%, center, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

mw {
  background-clip: content-box, padding-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box;
  background-position: 5px bottom, left top, left top;
  background-repeat: repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto;
}

nw {
  background-clip: content-box, content-box, border-box, border-box, content-box, content-box, padding-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none, none, none, none /* FIXME */, none, none /* FIXME */;
  background-origin: content-box, content-box, padding-box, padding-box, content-box, content-box, padding-box;
  background-position: left top, 5px bottom, left top, left top, left top, 5px bottom, center;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto;
}

ow {
  background-clip: padding-box, padding-box, content-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: content-box, content-box, border-box;
  background-position: left top, 10%, left top;
  background-repeat: repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, auto;
}

pw {
  background-clip: padding-box, border-box, border-box, border-box, padding-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, padding-box, border-box, content-box, border-box, content-box, content-box;
  background-position: 5px bottom, 10%, left top, left top, left top, center, left top, left top;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, cover, auto, auto, auto;
}

qw {
  background-clip: padding-box, border-box;
  background-color: currentColor;
  background-image: none, none;
  background-origin: content-box, border-box;
  background-position: 5px bottom, left top;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

rw {
  background-clip: border-box, border-box, content-box;
  background-color: currentColor;
  background-image: none, none, none;
  background-origin: padding-box, border-box, content-box;
  background-position: left top, left top, left top;
  background-repeat: round no-repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

sw {
  background-clip: content-box, border-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, border-box;
  background-position: left top, left top;
  background-repeat: repeat, repeat;
  background-size: auto, 10%;
}

tw {
  background-clip: border-box, padding-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, padding-box;
  background-position: left top, center, center;
  background-repeat: repeat, repeat, round no-repeat;
  background-size: auto, auto, auto;
}

uw {
  background-clip: border-box, border-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, none /* FIXME */, none /* FIXME */, none /* FIXME */;
  background-origin: border-box, border-box, border-box, padding-box;
  background-position: left top, 10%, center, 10%;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto;
}

vw {
  background-clip: padding-box, padding-box, border-box, content-box, border-box, content-box, padding-box, border-box, border-box, content-box, padding-box;
  background-color: @home;
  background-image: none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, border-box, content-box, content-box, content-box, border-box, padding-box, padding-box, content-box, padding-box;
  background-position: 5px bottom, center, left top, 10%, left top, 10%, 10%, left top, left top, 5px bottom, 10%;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: cover, auto, auto, 10%, auto, auto, auto, auto, auto, 10%, auto;
}

ww {
  background-clip: padding-box, padding-box, border-box, content-box, content-box, content-box, border-box, content-box, border-box, padding-box, padding-box, content-box, border-box, border-box, padding-box, content-box, border-box, padding-box, border-box, border-box, padding-box, border-box, border-box, padding-box, content-box, border-box, border-box, border-box, border-box, padding-box, padding-box, border-box, padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none /* FIXME */, none /* FIXME */, none, none, none /* FIXME */, none, none, none, none /* FIXME */, none /* FIXME */, none, none;
  background-origin: padding-box, padding-box, padding-box, content-box, content-box, border-box, border-box, content-box, border-box, padding-box, padding-box, content-box, border-box, padding-box, border-box, content-box, border-box, padding-box, border-box, border-box, padding-box, padding-box, padding-box, padding-box, content-box, border-box, padding-box, border-box, border-box, padding-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, 10%, 10%, 10%, left top, 10%, 5px bottom, left top, 5px bottom, 5px bottom, left top, 10%, 10%, left top, 10%, left top, left top, left top, 5px bottom, 10%, left top, 10%, center, 5px bottom, center, left top, left top, left top, 5px bottom, left top, left top, 10%, 5px bottom;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, cover, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto;
}

xw {
  background-clip: border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, none;
  background-origin: padding-box, padding-box;
  background-position: left top, center;
  background-repeat: repeat, repeat;
  background-size: auto, 10%;
}

yw {
  background-clip: content-box, padding-box, border-box, content-box, content-box, border-box;
  background-color: @home;
  background-image: none, none, none, none, none, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, content-box, content-box, padding-box;
  background-position: left top, 5px bottom, left top, center, left top, left top;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, 10%, auto, auto, auto, auto;
}

zw {
  background-clip: content-box, padding-box;
  background-color: initial;
  background-image: none, none;
  background-origin: content-box, content-box;
  background-position: left top, left top;
  background-repeat: round no-repeat, round no-repeat;
  background-size: cover, auto;
}

ax {
  background-clip: content-box, border-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box, content-box;
  background-position: 5px bottom, left top, center;
  background-repeat: round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto;
}

bx {
  background-clip: border-box, content-box;
  background-color: initial;
  background-image: none, none;
  background-origin: padding-box, border-box;
  background-position: center, left top;
  background-repeat: repeat, round no-repeat;
  background-size: 10%, auto;
}

cx {
  background-clip: padding-box, border-box;
  background-color: initial;
  background-image: none, none;
  background-origin: padding-box, padding-box;
  background-position: 10%, center;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

dx {
  background-clip: content-box, content-box, content-box, content-box, padding-box, border-box, padding-box, border-box, content-box, content-box, border-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box, border-box, content-box, padding-box, padding-box, padding-box, padding-box, content-box, content-box, border-box;
  background-position: 5px bottom, left top, left top, left top, 5px bottom, 5px bottom, 5px bottom, 5px bottom, left top, left top, left top;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: cover, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto;
}

ex {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: border-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

fx {
  background-clip: content-box;
  background-color: @home;
  background-image: none /* FIXME */;
  background-origin: border-box;
  background-position: 10%;
  background-repeat: repeat;
  background-size: auto;
}

gx {
  background-clip: padding-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none;
  background-origin: content-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

hx {
  background-clip: border-box, border-box, padding-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none;
  background-origin: border-box, border-box, padding-box;
  background-position: 5px bottom, 5px bottom, center;
  background-repeat: round no-repeat, round no-repeat, repeat;
  background-size: cover, auto, auto;
}

ix {
  background-clip: border-box, content-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none, none /* FIXME */, none;
  background-origin: border-box, content-box, content-box;
  background-position: left top, 10%, 5px bottom;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

jx {
  background-clip: border-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: auto;
}

kx {
  background-clip: border-box;
  background-color: currentColor;
  background-image: none;
  background-origin: padding-box;
  background-position: 5px bottom;
  background-repeat: repeat;
  background-size: 10%;
}

lx {
  background-clip: border-box, content-box, border-box, content-box, padding-box;
  background-color: initial;
  background-image: none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: border-box, content-box, border-box, content-box, padding-box;
  background-position: left top, center, 5px bottom, left top, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, 10%;
}

mx {
  background-clip: padding-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box;
  background-position: 5px bottom, 5px bottom;
  background-repeat: round no-repeat, repeat;
  background-size: auto, auto;
}

nx {
  background-clip: content-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none;
  background-origin: content-box, padding-box;
  background-position: center, left top;
  background-repeat: round no-repeat, repeat;
  background-size: 10%, auto;
}

ox {
  background-clip: border-box;
  background-color: rgb(255,0,0);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: 10%;
}

px {
  background-clip: border-box, border-box, content-box, content-box;
  background-color: @home;
  background-image: none /* FIXME */, none, none /* FIXME */, none;
  background-origin: border-box, padding-box, content-box, content-box;
  background-position: 10%, center, center, center;
  background-repeat: round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto;
}

qx {
  background-clip: border-box, border-box, content-box, content-box;
  background-color: currentColor;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, border-box, content-box, content-box;
  background-position: left top, 5px bottom, 10%, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

rx {
  background-clip: content-box, padding-box, padding-box, content-box, padding-box, border-box, content-box, padding-box, content-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, padding-box, padding-box, border-box, border-box, padding-box, border-box, padding-box;
  background-position: 5px bottom, left top, left top, 5px bottom, left top, left top, left top, left top, 5px bottom, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat;
  background-size: auto, auto, 10%, auto, auto, 10%, auto, auto, auto, 10%;
}

sx {
  background-clip: border-box, padding-box, border-box, border-box, padding-box, content-box, padding-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none;
  background-origin: border-box, content-box, border-box, border-box, padding-box, border-box, padding-box, border-box;
  background-position: left top, center, center, 5px bottom, left top, 10%, 10%, 10%;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, cover, 10%, auto, auto, auto, auto, auto;
}

tx {
  background-clip: content-box, padding-box, border-box, padding-box, content-box, content-box, padding-box, border-box, border-box, padding-box, border-box, content-box, padding-box;
  background-color: @home;
  background-image: none, none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: padding-box, padding-box, border-box, padding-box, content-box, padding-box, padding-box, border-box, border-box, content-box, padding-box, content-box, padding-box;
  background-position: center, left top, 5px bottom, center, left top, 10%, left top, left top, left top, 10%, left top, 5px bottom, 5px bottom;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, cover, auto, auto, auto, auto, auto, auto, auto, cover, auto, cover, 10%;
}

ux {
  background-clip: border-box, border-box, border-box, border-box, padding-box, border-box, padding-box, border-box, content-box, border-box, border-box, border-box, border-box, border-box, padding-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none;
  background-origin: border-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box, content-box, border-box, padding-box, border-box, border-box, border-box, padding-box, padding-box;
  background-position: center, 5px bottom, center, 10%, 5px bottom, center, 5px bottom, center, left top, left top, left top, center, left top, left top, center, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, 10%, auto, auto, 10%, auto, auto, 10%, auto, 10%;
}

vx {
  background-clip: border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box;
  background-position: center;
  background-repeat: round no-repeat;
  background-size: auto;
}

wx {
  background-clip: padding-box, border-box, content-box;
  background-color: currentColor;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, border-box, content-box;
  background-position: center, left top, left top;
  background-repeat: repeat, repeat, round no-repeat;
  background-size: auto, auto, auto;
}

xx {
  background-clip: border-box, border-box, border-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, left top, left top, center;
  background-repeat: round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

yx {
  background-clip: padding-box, border-box, content-box, padding-box, border-box, border-box, border-box;
  background-color: currentColor;
  background-image: none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none;
  background-origin: padding-box, padding-box, border-box, padding-box, padding-box, content-box, border-box;
  background-position: 10%, 5px bottom, 5px bottom, 5px bottom, 10%, left top, left top;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: cover, auto, cover, auto, 10%, auto, auto;
}

zx {
  background-clip: padding-box, border-box, padding-box;
  background-color: currentColor;
  background-image: none /* FIXME */, none /* FIXME */, none;
  background-origin: content-box, padding-box, padding-box;
  background-position: left top, 5px bottom, 10%;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, 10%;
}

ay {
  background-clip: padding-box, content-box, border-box;
  background-color: initial;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, padding-box, padding-box;
  background-position: left top, 5px bottom, left top;
  background-repeat: round no-repeat, round no-repeat, round no-repeat;
  background-size: cover, auto, auto;
}

by {
  background-clip: padding-box, border-box, padding-box, content-box;
  background-color: currentColor;
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: padding-box, padding-box, border-box, padding-box;
  background-position: 5px bottom, 10%, left top, center;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat;
  background-size: 10%, auto, auto, auto;
}

cy {
  background-clip: border-box, content-box;
  background-color: @home;
  background-image: none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, content-box;
  background-position: 10%, left top;
  background-repeat: repeat, repeat;
  background-size: 10%, auto;
}

dy {
  background-clip: content-box, content-box, padding-box, padding-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none;
  background-origin: content-box, content-box, padding-box, padding-box;
  background-position: left top, left top, left top, 10%;
  background-repeat: repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto;
}

ey {
  background-clip: border-box, padding-box, content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, content-box, padding-box;
  background-position: left top, 10%, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, 10%;
}

fy {
  background-clip: border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, none;
  background-origin: padding-box, border-box;
  background-position: left top, center;
  background-repeat: repeat, round no-repeat;
  background-size: auto, auto;
}

gy {
  background-clip: content-box, border-box, content-box, padding-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, padding-box, content-box, border-box;
  background-position: 5px bottom, center, 10%, left top;
  background-repeat: repeat, repeat, repeat, repeat;
  background-size: cover, 10%, auto, auto;
}

hy {
  background-clip: content-box;
  background-color: currentColor;
  background-image: none;
  background-origin: padding-box;
  background-position: 10%;
  background-repeat: round no-repeat;
  background-size: auto;
}

iy {
  background-clip: border-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

jy {
  background-clip: border-box, padding-box, padding-box, border-box, content-box, content-box, border-box, border-box, border-box, content-box, padding-box, content-box, border-box, content-box, content-box, content-box, padding-box, padding-box, padding-box, padding-box, padding-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */;
  background-origin: padding-box, padding-box, border-box, border-box, content-box, content-box, padding-box, content-box, padding-box, content-box, padding-box, content-box, border-box, content-box, content-box, content-box, padding-box, content-box, padding-box, border-box, content-box;
  background-position: 5px bottom, 10%, center, center, 10%, 5px bottom, left top, left top, 10%, left top, center, 5px bottom, left top, center, center, left top, left top, center, 5px bottom, 5px bottom, center;
  background-repeat: repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat;
  background-size: auto, cover, auto, cover, auto, 10%, auto, auto, auto, auto, 10%, auto, 10%, auto, auto, auto, auto, 10%, auto, auto, cover;
}

ky {
  background-clip: content-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */;
  background-origin: content-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

ly {
  background-clip: border-box, border-box;
  background-color: initial;
  background-image: none, none;
  background-origin: border-box, border-box;
  background-position: center, 10%;
  background-repeat: repeat, repeat;
  background-size: 10%, 10%;
}

my {
  background-clip: border-box, content-box, padding-box, padding-box, border-box;
  background-color: initial;
  background-image: none, none, none, none /* FIXME */, none;
  background-origin: padding-box, content-box, padding-box, padding-box, border-box;
  background-position: center, left top, center, left top, left top;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto;
}

ny {
  background-clip: border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

oy {
  background-clip: padding-box, border-box, padding-box, border-box, padding-box, border-box, padding-box, border-box, content-box, border-box, content-box, padding-box, content-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, none, none, none, none /* FIXME */, none, none /* FIXME */, none, none, none /* FIXME */, none, none;
  background-origin: padding-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box, content-box, border-box, border-box, border-box, content-box;
  background-position: 10%, left top, left top, 10%, center, center, left top, center, left top, 5px bottom, left top, left top, center;
  background-repeat: repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto;
}

py {
  background-clip: content-box, border-box, border-box, border-box, border-box, border-box, border-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */, none, none, none /* FIXME */;
  background-origin: content-box, border-box, padding-box, content-box, border-box, border-box, padding-box, border-box, padding-box;
  background-position: left top, center, 10%, left top, left top, center, left top, left top, left top;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, 10%;
}

qy {
  background-clip: border-box, content-box, border-box, border-box, border-box, content-box, border-box, padding-box, content-box, content-box;
  background-color: @home;
  background-image: none, none, none, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */;
  background-origin: border-box, border-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box, content-box, content-box;
  background-position: center, left top, 10%, center, left top, left top, left top, center, left top, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, cover, 10%, auto, cover, auto, cover, auto, cover;
}

ry {
  background-clip: padding-box, border-box, padding-box, content-box, border-box;
  background-color: initial;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: padding-box, border-box, padding-box, content-box, border-box;
  background-position: center, center, left top, left top, left top;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, 10%, auto, auto, auto;
}

sy {
  background-clip: border-box, border-box, content-box, padding-box, padding-box, border-box, content-box, padding-box, content-box, border-box, border-box, padding-box, content-box, border-box, content-box, border-box, content-box, padding-box;
  background-color: currentColor;
  background-image: none, none, none, none, none, none, none, none /* FIXME */, none, none, none, none /* FIXME */, none /* FIXME */, none, none, none, none, none;
  background-origin: padding-box, border-box, content-box, padding-box, padding-box, padding-box, content-box, padding-box, border-box, border-box, padding-box, padding-box, border-box, content-box, padding-box, padding-box, content-box, padding-box;
  background-position: left top, 5px bottom, center, left top, left top, 10%, left top, left top, left top, 10%, left top, left top, left top, center, 10%, left top, 5px bottom, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: 10%, auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto, auto, auto, cover, auto, auto, cover, auto;
}

ty {
  background-clip: content-box, content-box, padding-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, content-box, padding-box, padding-box, border-box, border-box;
  background-position: 5px bottom, left top, 5px bottom, left top, 5px bottom, 10%;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto;
}

uy {
  background-clip: content-box, border-box, border-box, border-box, padding-box, border-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none /* FIXME */;
  background-origin: padding-box, border-box, padding-box, border-box, padding-box, padding-box, border-box, border-box;
  background-position: left top, left top, 10%, left top, 5px bottom, 10%, 10%, 5px bottom;
  background-repeat: repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, 10%;
}

vy {
  background-clip: border-box, padding-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box, border-box, border-box, content-box, border-box, padding-box, border-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, padding-box, padding-box, padding-box, content-box, border-box, padding-box, border-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box;
  background-position: 5px bottom, left top, 10%, 5px bottom, 10%, 10%, 5px bottom, 5px bottom, left top, 10%, 5px bottom, 10%, 5px bottom, left top, 5px bottom, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, 10%, auto, auto, auto;
}

wy {
  background-clip: border-box, border-box, border-box, border-box, padding-box, padding-box, border-box, padding-box, border-box, border-box, border-box, content-box, content-box, border-box, content-box, border-box, padding-box;
  background-color: @home;
  background-image: none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, border-box, border-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, content-box, padding-box, content-box, padding-box, content-box;
  background-position: 5px bottom, 10%, 10%, 10%, 5px bottom, left top, left top, 10%, left top, left top, left top, 5px bottom, 10%, 10%, left top, left top, center;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: cover, auto, auto, cover, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, cover;
}

xy {
  background-clip: content-box, padding-box, border-box, padding-box, border-box, border-box, content-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none, none, none, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, content-box, padding-box, padding-box, content-box, padding-box, border-box, padding-box;
  background-position: center, center, left top, left top, left top, 10%, left top, 5px bottom, left top, left top;
  background-repeat: repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, auto, cover, auto, auto;
}

yy {
  background-clip: border-box, border-box, border-box, border-box, border-box, border-box, padding-box, border-box, content-box, padding-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: border-box, padding-box, border-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box;
  background-position: center, left top, 5px bottom, left top, left top, left top, 5px bottom, 10%, left top, center, left top, 10%, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, 10%, auto, cover, auto, auto, auto, auto, auto, auto, auto;
}

zy {
  background-clip: border-box, border-box, border-box, border-box, padding-box, border-box, padding-box, border-box, content-box, padding-box, padding-box, border-box, content-box, content-box;
  background-color: @home;
  background-image: none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none, none /* FIXME */, none, none, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none;
  background-origin: padding-box, padding-box, border-box, padding-box, padding-box, border-box, content-box, padding-box, content-box, content-box, padding-box, border-box, border-box, content-box;
  background-position: left top, center, center, 5px bottom, left top, center, left top, 5px bottom, 5px bottom, left top, 5px bottom, 10%, 10%, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, round no-repeat, round no-repeat, repeat, repeat;
  background-size: auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, 10%, auto;
}

az {
  background-clip: padding-box, border-box, border-box, border-box;
  background-color: @home;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: border-box, padding-box, content-box, content-box;
  background-position: left top, 5px bottom, 10%, 5px bottom;
  background-repeat: repeat, round no-repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto;
}

bz {
  background-clip: border-box, content-box, border-box, padding-box, border-box, content-box, border-box, border-box, border-box, padding-box, border-box, border-box, border-box, border-box, content-box, padding-box, border-box, border-box, padding-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: none, none, none, none, none /* FIXME */, none, none, none /* FIXME */, none, none /* FIXME */, none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, content-box, padding-box, padding-box, border-box, padding-box, padding-box, padding-box, padding-box, padding-box, border-box, padding-box, border-box, padding-box, content-box, padding-box, padding-box, padding-box, border-box, padding-box, padding-box, border-box;
  background-position: left top, left top, left top, left top, 10%, center, 5px bottom, left top, 5px bottom, 10%, 10%, 10%, left top, 5px bottom, 10%, left top, left top, 5px bottom, 5px bottom, left top, left top, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, cover, auto, auto, auto, 10%, auto, cover, auto, auto, auto, auto, auto;
}

cz {
  background-clip: content-box, border-box, border-box, border-box, border-box, border-box, padding-box, content-box, border-box;
  background-color: rgb(255,0,0);
  background-image: none, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none /* FIXME */;
  background-origin: content-box, padding-box, padding-box, padding-box, padding-box, padding-box, padding-box, content-box, padding-box;
  background-position: 10%, 5px bottom, left top, 5px bottom, left top, left top, center, left top, 10%;
  background-repeat: repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, 10%, auto, auto, auto, 10%, 10%;
}

dz {
  background-clip: border-box, content-box, border-box, border-box, border-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none, none, none, none /* FIXME */, none, none /* FIXME */;
  background-origin: border-box, border-box, padding-box, border-box, border-box, content-box;
  background-position: center, left top, 10%, center, 5px bottom, 5px bottom;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: cover, auto, auto, cover, auto, 10%;
}

ez {
  background-clip: border-box, border-box, padding-box, padding-box;
  background-color: rgb(170,221,255);
  background-image: none /* FIXME */, none /* FIXME */, none, none;
  background-origin: padding-box, padding-box, content-box, padding-box;
  background-position: 10%, 10%, 5px bottom, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, auto, auto;
}

fz {
  background-clip: padding-box, padding-box, border-box;
  background-color: @home;
  background-image: none, none, none;
  background-origin: padding-box, padding-box, padding-box;
  background-position: 5px bottom, center, 5px bottom;
  background-repeat: round no-repeat, repeat, round no-repeat;
  background-size: auto, auto, auto;
}

gz {
  background-clip: padding-box, border-box, border-box, content-box;
  background-color: initial;
  background-image: none /* FIXME */, none, none /* FIXME */, none;
  background-origin: padding-box, padding-box, padding-box, content-box;
  background-position: 10%, center, 10%, 5px bottom;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat;
  background-size: auto, cover, auto, 10%;
}

hz {
  background-clip: border-box, padding-box, content-box, padding-box, border-box;
  background-color: rgb(170,221,255);
  background-image: none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: border-box, content-box, padding-box, border-box, border-box;
  background-position: center, 10%, 10%, center, 5px bottom;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, round no-repeat;
  background-size: cover, auto, auto, cover, auto;
}

iz {
  background-clip: content-box, border-box, border-box, padding-box, border-box, content-box, border-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none;
  background-origin: content-box, border-box, padding-box, border-box, padding-box, content-box, padding-box, border-box;
  background-position: 10%, left top, center, center, left top, center, left top, left top;
  background-repeat: round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, auto, auto, cover, auto, auto, auto;
}

jz {
  background-clip: content-box, content-box, border-box;
  background-color: rgba(255,192,203,0.5);
  background-image: none /* FIXME */, none /* FIXME */, none;
  background-origin: padding-box, content-box, padding-box;
  background-position: left top, 5px bottom, center;
  background-repeat: repeat, round no-repeat, repeat;
  background-size: auto, auto, auto;
}

kz {
  background-clip: padding-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

lz {
  background-clip: padding-box;
  background-color: @home;
  background-image: none;
  background-origin: padding-box;
  background-position: center;
  background-repeat: repeat;
  background-size: auto;
}

mz {
  background-clip: padding-box, padding-box, content-box, border-box, content-box, border-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, padding-box, content-box, padding-box, border-box, padding-box;
  background-position: left top, left top, 5px bottom, 10%, left top, 10%;
  background-repeat: repeat, round no-repeat, repeat, repeat, repeat, repeat;
  background-size: auto, auto, cover, auto, auto, auto;
}

nz {
  background-clip: border-box;
  background-color: rgb(170,221,255);
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: repeat;
  background-size: auto;
}

oz {
  background-clip: padding-box, padding-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: content-box, padding-box;
  background-position: 10%, 5px bottom;
  background-repeat: repeat, round no-repeat;
  background-size: 10%, auto;
}

pz {
  background-clip: content-box, border-box;
  background-color: @home;
  background-image: none, none /* FIXME */;
  background-origin: content-box, border-box;
  background-position: 10%, left top;
  background-repeat: repeat, round no-repeat;
  background-size: cover, auto;
}

qz {
  background-clip: border-box, border-box, padding-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none;
  background-origin: padding-box, border-box, padding-box, border-box;
  background-position: 5px bottom, 5px bottom, left top, 10%;
  background-repeat: repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto;
}

rz {
  background-clip: padding-box, content-box, border-box, border-box, border-box, border-box, border-box;
  background-color: currentColor;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */, none, none, none;
  background-origin: padding-box, border-box, border-box, border-box, padding-box, padding-box, border-box;
  background-position: 5px bottom, left top, center, center, center, center, left top;
  background-repeat: repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, 10%, auto, auto;
}

sz {
  background-clip: content-box;
  background-color: @home;
  background-image: none /* FIXME */;
  background-origin: content-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

tz {
  background-clip: padding-box, padding-box, content-box, border-box, content-box, padding-box, padding-box, border-box, border-box, border-box, padding-box, border-box, border-box, border-box, border-box;
  background-color: initial;
  background-image: none, none, none /* FIXME */, none /* FIXME */, none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none;
  background-origin: padding-box, padding-box, content-box, padding-box, content-box, padding-box, padding-box, border-box, border-box, padding-box, padding-box, border-box, padding-box, border-box, border-box;
  background-position: left top, left top, center, left top, center, 10%, left top, left top, center, 5px bottom, left top, center, center, left top, 10%;
  background-repeat: round no-repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, auto, 10%, auto, auto, auto, auto, auto, auto, cover, auto, auto, cover, cover, auto;
}

uz {
  background-clip: content-box, content-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */;
  background-origin: content-box, padding-box;
  background-position: 10%, left top;
  background-repeat: repeat, repeat;
  background-size: 10%, auto;
}

vz {
  background-clip: border-box, border-box, content-box, border-box, border-box;
  background-color: currentColor;
  background-image: none, none /* FIXME */, none, none, none /* FIXME */;
  background-origin: border-box, padding-box, content-box, padding-box, content-box;
  background-position: left top, 10%, 5px bottom, center, 5px bottom;
  background-repeat: round no-repeat, repeat, repeat, round no-repeat, repeat;
  background-size: auto, auto, auto, auto, auto;
}

wz {
  background-clip: padding-box, border-box;
  background-color: initial;
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: padding-box, border-box;
  background-position: 5px bottom, center;
  background-repeat: repeat, repeat;
  background-size: auto, auto;
}

xz {
  background-clip: content-box;
  background-color: initial;
  background-image: none;
  background-origin: padding-box;
  background-position: left top;
  background-repeat: round no-repeat;
  background-size: auto;
}

yz {
  background-clip: content-box, border-box, content-box;
  background-color: rgb(170,221,255);
  background-image: linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128));
  background-origin: content-box, border-box, content-box;
  background-position: left top, 5px bottom, 5px bottom;
  background-repeat: repeat, repeat, repeat;
  background-size: auto, auto, auto;
}

zz {
  background-clip: border-box, border-box, padding-box, border-box, border-box, border-box, padding-box, border-box, content-box, border-box, content-box, content-box, padding-box, border-box, content-box, content-box, border-box, content-box, border-box, padding-box, border-box, border-box, border-box, border-box, border-box, border-box, padding-box, content-box, border-box, border-box;
  background-color: initial;
  background-image: none, none /* FIXME */, none, none, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), none /* FIXME */, none, none /* FIXME */, none /* FIXME */, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, linear-gradient(rgb(255,192,203), rgb(128,0,128)), linear-gradient(rgb(255,192,203), rgb(128,0,128)), none, none, none /* FIXME */;
  background-origin: padding-box, padding-box, padding-box, border-box, border-box, border-box, border-box, border-box, content-box, padding-box, padding-box, content-box, padding-box, padding-box, content-box, content-box, content-box, content-box, padding-box, content-box, padding-box, content-box, border-box, padding-box, border-box, padding-box, border-box, content-box, padding-box, padding-box;
  background-position: center, left top, 5px bottom, left top, 5px bottom, left top, center, left top, 10%, center, left top, 10%, center, left top, center, left top, 5px bottom, left top, left top, center, 10%, center, 5px bottom, left top, center, left top, left top, 10%, left top, left top;
  background-repeat: repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, repeat, repeat, repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, round no-repeat, round no-repeat, repeat, round no-repeat, repeat, round no-repeat, repeat, repeat, round no-repeat, repeat, repeat;
  background-size: auto, cover, auto, 10%, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, auto, cover, auto, auto, auto, auto, cover, auto, auto;
}
